@@ -1,6 +1,3 @@
-#include <stdio.h>
-#include <stdio.h>
-#include <sys/stat.h>
 #include "owl.h"
 
 /**********************************************************************/
@@ -112,11 +109,13 @@ void owl_aim_init(void)
      
 }
 
-void owl_aim_send_nop(owl_timer *t, void *data) {
-    if(owl_global_is_doaimevents(&g)) {
-        aim_session_t *sess = owl_global_get_aimsess(&g);
-        aim_flap_nop(sess, aim_getconn_type(sess, AIM_CONN_TYPE_BOS));
-    }
+gboolean owl_aim_send_nop(gpointer data) {
+  owl_global *g = data;
+  if (owl_global_is_doaimevents(g)) {
+    aim_session_t *sess = owl_global_get_aimsess(g);
+    aim_flap_nop(sess, aim_getconn_type(sess, AIM_CONN_TYPE_BOS));
+  }
+  return TRUE;
 }
 
 
@@ -133,12 +132,11 @@ int owl_aim_login(const char *screenname, const char *password)
   aim_tx_setenqueue(sess, AIM_TX_IMMEDIATE, NULL);
   
   /* this will leak, I know and just don't care right now */
-  priv=owl_malloc(sizeof(struct owlfaim_priv));
-  memset(priv, 0, sizeof(struct owlfaim_priv));
+  priv=g_new0(struct owlfaim_priv, 1);
 
-  priv->screenname = owl_strdup(screenname);
-  priv->password = owl_strdup(password);
-  priv->server = owl_strdup(FAIM_LOGIN_SERVER);
+  priv->screenname = g_strdup(screenname);
+  priv->password = g_strdup(password);
+  priv->server = g_strdup(FAIM_LOGIN_SERVER);
   sess->aux_data = priv;
 
   conn=aim_newconn(sess, AIM_CONN_TYPE_AUTH, priv->server ? priv->server : FAIM_LOGIN_SERVER);
@@ -183,14 +181,16 @@ int owl_aim_login(const char *screenname, const char *password)
   aim_request_login(sess, conn, screenname);
   owl_function_debugmsg("owl_aim_login: connecting");
 
-  g.aim_nop_timer = owl_select_add_timer(30, 30, owl_aim_send_nop, NULL, NULL);
+  g.aim_nop_timer = g_timeout_add_seconds(30, owl_aim_send_nop, &g);
 
   return(0);
 }
 
-static void owl_aim_unset_ignorelogin(owl_timer *t, void *data)
+static gboolean owl_aim_unset_ignorelogin(void *data)
 {
-    owl_global_unset_ignore_aimlogin(&g);
+  owl_global *g = data;
+  owl_global_unset_ignore_aimlogin(g);
+  return FALSE;  /* only run once. */
 }
 
 /* stuff to run once login has been successful */
@@ -201,16 +201,16 @@ void owl_aim_successful_login(const char *screenname)
   owl_global_set_aimloggedin(&g, screenname);
   owl_global_set_doaimevents(&g); /* this should already be on */
   owl_function_makemsg("%s logged in", screenname);
-  buff=owl_sprintf("Logged in to AIM as %s", screenname);
+  buff=g_strdup_printf("Logged in to AIM as %s", screenname);
   owl_function_adminmsg("", buff);
-  owl_free(buff);
+  g_free(buff);
 
   owl_function_debugmsg("Successful AIM login for %s", screenname);
 
   /* start the ingorelogin timer */
   owl_global_set_ignore_aimlogin(&g);
-  owl_select_add_timer(owl_global_get_aim_ignorelogin_timer(&g),
-                       0, owl_aim_unset_ignorelogin, NULL, NULL);
+  g_timeout_add_seconds(owl_global_get_aim_ignorelogin_timer(&g),
+                        owl_aim_unset_ignorelogin, &g);
 
   /* aim_ssi_setpresence(owl_global_get_aimsess(&g), 0x00000400); */
   /* aim_bos_setidle(owl_global_get_aimsess(&g), owl_global_get_bosconn(&g), 5000); */
@@ -224,7 +224,10 @@ void owl_aim_logout(void)
   if (owl_global_is_aimloggedin(&g)) owl_function_adminmsg("", "Logged out of AIM");
   owl_global_set_aimnologgedin(&g);
   owl_global_set_no_doaimevents(&g);
-  owl_select_remove_timer(g.aim_nop_timer);
+  if (g.aim_nop_timer) {
+    g_source_remove(g.aim_nop_timer);
+    g.aim_nop_timer = 0;
+  }
 }
 
 void owl_aim_logged_out(void)
@@ -243,7 +246,10 @@ void owl_aim_login_error(const char *message)
   owl_function_beep();
   owl_global_set_aimnologgedin(&g);
   owl_global_set_no_doaimevents(&g);
-  owl_select_remove_timer(g.aim_nop_timer);
+  if (g.aim_nop_timer) {
+    g_source_remove(g.aim_nop_timer);
+    g.aim_nop_timer = 0;
+  }
 }
 
 /*
@@ -291,7 +297,7 @@ static int owl_aim_do_send(const char *to, const char *msg, int flags)
 
   ret=aim_im_sendch1_ext(owl_global_get_aimsess(&g), &args);
 
-  owl_free(encoded);
+  g_free(encoded);
 
   return(ret);
 }
@@ -345,7 +351,7 @@ int owl_aim_set_awaymsg(const char *msg)
   char *foo;
   /* there is a max away message lentgh we should check against */
 
-  foo=owl_strdup(msg);
+  foo=g_strdup(msg);
   len=strlen(foo);
   if (len>500) {
     foo[500]='\0';
@@ -355,7 +361,7 @@ int owl_aim_set_awaymsg(const char *msg)
   aim_locate_setprofile(owl_global_get_aimsess(&g),
 			NULL, NULL, 0,
 			"us-ascii", foo, len);
-  owl_free(foo);
+  g_free(foo);
 
   /*
   aim_bos_setprofile(owl_global_get_aimsess(&g),
@@ -427,13 +433,13 @@ int owl_aim_chat_sendmsg(const char *chatroom, const char *msg)
 }
 
 /* caller must free the return */
-char *owl_aim_normalize_screenname(const char *in)
+CALLER_OWN char *owl_aim_normalize_screenname(const char *in)
 {
   char *out;
   int i, j, k;
 
   j=strlen(in);
-  out=owl_malloc(j+30);
+  out=g_new(char, j+30);
   k=0;
   for (i=0; i<j; i++) {
     if (in[i]!=' ') {
@@ -445,15 +451,13 @@ char *owl_aim_normalize_screenname(const char *in)
   return(out);
 }
 
-int owl_aim_process_events(void)
+int owl_aim_process_events(aim_session_t *aimsess)
 {
-  aim_session_t *aimsess;
   aim_conn_t *waitingconn = NULL;
   struct timeval tv;
   int selstat = 0;
   struct owlfaim_priv *priv;
 
-  aimsess=owl_global_get_aimsess(&g);
   priv = aimsess->aux_data;
 
   /* do a select without blocking */
@@ -1024,7 +1028,7 @@ static int faimtest_parse_incoming_im_chan1(aim_session_t *sess, aim_conn_t *con
   char *realmsg = NULL;
 
   if (!args->msg) {
-    realmsg = owl_strdup("");
+    realmsg = g_strdup("");
   } else if (args->icbmflags & AIM_IMFLAGS_UNICODE) {
     realmsg = g_convert(args->msg, args->msglen, "UTF-8", "UCS-2BE",
                         NULL, NULL, NULL);
@@ -1032,11 +1036,11 @@ static int faimtest_parse_incoming_im_chan1(aim_session_t *sess, aim_conn_t *con
     realmsg = g_convert(args->msg, args->msglen, "UTF-8", "ISO-8859-1",
                         NULL, NULL, NULL);
   } else {
-    realmsg = owl_strdup(args->msg);
+    realmsg = g_strdup(args->msg);
   }
 
   if (!realmsg) {
-    realmsg = owl_strdup("[Error decoding incoming IM]");
+    realmsg = g_strdup("[Error decoding incoming IM]");
   }
 
   owl_function_debugmsg("faimtest_parse_incoming_im_chan1: message from: %s", userinfo->sn?userinfo->sn:"");
@@ -1044,7 +1048,7 @@ static int faimtest_parse_incoming_im_chan1(aim_session_t *sess, aim_conn_t *con
   stripmsg=owl_text_htmlstrip(realmsg);
   wrapmsg=owl_text_wordwrap(stripmsg, 70);
   nz_screenname=owl_aim_normalize_screenname(userinfo->sn);
-  m=owl_malloc(sizeof(owl_message));
+  m=g_new(owl_message, 1);
   owl_message_create_aim(m,
 			 nz_screenname,
 			 owl_global_get_aim_screenname(&g),
@@ -1053,37 +1057,14 @@ static int faimtest_parse_incoming_im_chan1(aim_session_t *sess, aim_conn_t *con
 			 0);
   if (args->icbmflags & AIM_IMFLAGS_AWAY) owl_message_set_attribute(m, "isauto", "");
   owl_global_messagequeue_addmsg(&g, m);
-  owl_free(stripmsg);
-  owl_free(wrapmsg);
-  owl_free(nz_screenname);
+  g_free(stripmsg);
+  g_free(wrapmsg);
+  g_free(nz_screenname);
+  g_free(realmsg);
 
   return(1);
 
-  owl_function_debugmsg("faimtest_parse_incoming_im_chan1: icbm: message: %s\n", realmsg);
-  
-  if (args->icbmflags & AIM_IMFLAGS_MULTIPART) {
-    aim_mpmsg_section_t *sec;
-    int z;
-
-    owl_function_debugmsg("faimtest_parse_incoming_im_chan1: icbm: multipart: this message has %d parts\n", args->mpmsg.numparts);
-    
-    for (sec = args->mpmsg.parts, z = 0; sec; sec = sec->next, z++) {
-      if ((sec->charset == 0x0000) || (sec->charset == 0x0003) || (sec->charset == 0xffff)) {
-	owl_function_debugmsg("faimtest_parse_incoming_im_chan1: icbm: multipart:   part %d: charset 0x%04x, subset 0x%04x, msg = %s\n", z, sec->charset, sec->charsubset, sec->data);
-      } else {
-	owl_function_debugmsg("faimtest_parse_incoming_im_chan1: icbm: multipart:   part %d: charset 0x%04x, subset 0x%04x, binary or UNICODE data\n", z, sec->charset, sec->charsubset);
-      }
-    }
-  }
-  
-  if (args->icbmflags & AIM_IMFLAGS_HASICON) {
-    /* aim_send_im(sess, userinfo->sn, AIM_IMFLAGS_BUDDYREQ, "You have an icon"); */
-    owl_function_debugmsg("faimtest_parse_incoming_im_chan1: icbm: their icon: iconstamp = %ld, iconlen = 0x%08x, iconsum = 0x%04x\n", args->iconstamp, args->iconlen, args->iconsum);
-  }
-
-  owl_free(realmsg);
-
-  return(1);
+  /* TODO: Multipart? See history from before 1.8 release. */
 }
 
 /*
@@ -1194,7 +1175,7 @@ static int faimtest_parse_oncoming(aim_session_t *sess, aim_frame_t *fr, ...)
   
   owl_function_debugmsg("parse_oncoming sn: %s idle: %i", userinfo->sn, userinfo->idletime);
     
-  owl_free(nz_screenname);
+  g_free(nz_screenname);
   
   /*
     printf("%ld  %s is now online (flags: %04x = %s%s%s%s%s%s%s%s) (caps = %s = 0x%08lx)\n",
@@ -1226,7 +1207,7 @@ static int faimtest_parse_offgoing(aim_session_t *sess, aim_frame_t *fr, ...)
 
   nz_screenname=owl_aim_normalize_screenname(userinfo->sn);
   owl_buddylist_offgoing(owl_global_get_buddylist(&g), nz_screenname);
-  owl_free(nz_screenname);
+  g_free(nz_screenname);
 
   if (userinfo->present & AIM_USERINFO_PRESENT_IDLE) {
     owl_function_debugmsg("parse_offgoing sn: %s idle time %i", userinfo->sn, userinfo->idletime);
@@ -1431,7 +1412,7 @@ static int faimtest_parse_searchreply(aim_session_t *sess, aim_frame_t *fr, ...)
   va_list ap;
   const char *address, *SNs;
   int num, i;
-  owl_list list;
+  GPtrArray *list;
   
   va_start(ap, fr);
   address = va_arg(ap, const char *);
@@ -1439,16 +1420,16 @@ static int faimtest_parse_searchreply(aim_session_t *sess, aim_frame_t *fr, ...)
   SNs = va_arg(ap, const char *);
   va_end(ap);
 
-  owl_list_create(&list);
+  list = g_ptr_array_new();
   
   owl_function_debugmsg("faimtest_parse_searchreply: E-Mail Search Results for %s: ", address);
   for (i=0; i<num; i++) {
     owl_function_debugmsg("  %s", &SNs[i*(MAXSNLEN+1)]);
-    owl_list_append_element(&list, (void *)&SNs[i*(MAXSNLEN+1)]);
+    g_ptr_array_add(list, (void *)&SNs[i*(MAXSNLEN+1)]);
   }
-  owl_function_aimsearch_results(address, &list);
-  owl_list_cleanup(&list, NULL);
-  return(1);
+  owl_function_aimsearch_results(address, list);
+  g_ptr_array_free(list, true);
+  return 1;
 }
 
 static int faimtest_parse_searcherror(aim_session_t *sess, aim_frame_t *fr, ...)
@@ -1794,9 +1775,117 @@ void chat_redirect(aim_session_t *sess, struct aim_redirect_data *redir)
   return;	
 }
 
-void owl_process_aim(void)
+typedef struct _owl_aim_event_source { /*noproto*/
+  GSource source;
+  aim_session_t *sess;
+  GPtrArray *fds;
+} owl_aim_event_source;
+
+static void truncate_pollfd_list(owl_aim_event_source *event_source, int len)
 {
-  if (owl_global_is_doaimevents(&g)) {
-    owl_aim_process_events();
+  GPollFD *fd;
+  int i;
+  if (len < event_source->fds->len) {
+    owl_function_debugmsg("Truncating AIM PollFDs to %d, was %d", len, event_source->fds->len);
+    for (i = len; i < event_source->fds->len; i++) {
+      fd = event_source->fds->pdata[i];
+      g_source_remove_poll(&event_source->source, fd);
+      g_free(fd);
+    }
+    g_ptr_array_remove_range(event_source->fds, len, event_source->fds->len - len);
   }
 }
+
+static gboolean owl_aim_event_source_prepare(GSource *source, int *timeout)
+{
+  owl_aim_event_source *event_source = (owl_aim_event_source*)source;
+  aim_conn_t *cur;
+  GPollFD *fd;
+  int i;
+
+  /* AIM HACK:
+   *
+   *  The problem - I'm not sure where to hook into the owl/faim
+   *  interface to keep track of when the AIM socket(s) open and
+   *  close. In particular, the bosconn thing throws me off. So,
+   *  rather than register particular dispatchers for AIM, I look up
+   *  the relevant FDs and add them to select's watch lists, then
+   *  check for them individually before moving on to the other
+   *  dispatchers. --asedeno
+   */
+  i = 0;
+  for (cur = event_source->sess->connlist; cur; cur = cur->next) {
+    if (cur->fd != -1) {
+      /* Add new GPollFDs as necessary. */
+      if (i == event_source->fds->len) {
+	fd = g_new0(GPollFD, 1);
+	g_ptr_array_add(event_source->fds, fd);
+	g_source_add_poll(source, fd);
+        owl_function_debugmsg("Allocated new AIM PollFD, len = %d", event_source->fds->len);
+      }
+      fd = event_source->fds->pdata[i];
+      fd->fd = cur->fd;
+      fd->events = G_IO_IN | G_IO_HUP | G_IO_ERR;
+      if (cur->status & AIM_CONN_STATUS_INPROGRESS) {
+        /* AIM login requires checking writable sockets. See aim_select. */
+	fd->events |= G_IO_OUT;
+      }
+      i++;
+    }
+  }
+  /* If the number of GPollFDs went down, clean up. */
+  truncate_pollfd_list(event_source, i);
+
+  *timeout = -1;
+  return FALSE;
+}
+
+static gboolean owl_aim_event_source_check(GSource *source)
+{
+  owl_aim_event_source *event_source = (owl_aim_event_source*)source;
+  int i;
+
+  for (i = 0; i < event_source->fds->len; i++) {
+    GPollFD *fd = event_source->fds->pdata[i];
+    if (fd->revents & fd->events)
+      return TRUE;
+  }
+  return FALSE;
+}
+
+static gboolean owl_aim_event_source_dispatch(GSource *source, GSourceFunc callback, gpointer user_data)
+{
+  owl_aim_event_source *event_source = (owl_aim_event_source*)source;
+  owl_aim_process_events(event_source->sess);
+  return TRUE;
+}
+
+static void owl_aim_event_source_finalize(GSource *source)
+{
+  owl_aim_event_source *event_source = (owl_aim_event_source*)source;
+  /* Don't remove the GPollFDs. We are being finalized, so they'll be removed
+   * for us. Moreover, glib will fire asserts if g_source_remove_poll is called
+   * on a source which has been destroyed (which occurs when g_source_remove is
+   * called on it). */
+  owl_ptr_array_free(event_source->fds, g_free);
+}
+
+static GSourceFuncs aim_event_funcs = {
+  owl_aim_event_source_prepare,
+  owl_aim_event_source_check,
+  owl_aim_event_source_dispatch,
+  owl_aim_event_source_finalize,
+};
+
+GSource *owl_aim_event_source_new(aim_session_t *sess)
+{
+  GSource *source;
+  owl_aim_event_source *event_source;
+
+  source = g_source_new(&aim_event_funcs, sizeof(owl_aim_event_source));
+  event_source = (owl_aim_event_source *)source;
+  event_source->sess = sess;
+  /* TODO: When we depend on glib 2.22+, use g_ptr_array_new_with_free_func. */
+  event_source->fds = g_ptr_array_new();
+  return source;
+}
