@@ -1,49 +1,38 @@
-#include <stdio.h>
-#include <unistd.h>
-#include <stdlib.h>
-#include <string.h>
-#include <netdb.h>
-#include <termios.h>
-#include <sys/ioctl.h>
-#include <time.h>
 #include "owl.h"
+#include <stdio.h>
+#include <sys/ioctl.h>
 
-#ifndef MAXHOSTNAMELEN
-#define MAXHOSTNAMELEN 256
-#endif
+static void _owl_global_init_windows(owl_global *g);
 
 void owl_global_init(owl_global *g) {
-  struct hostent *hent;
-  char hostname[MAXHOSTNAMELEN];
   char *cd;
+  const char *homedir;
 
-  g->malloced=0;
-  g->freed=0;
+  g_type_init();
+  g_thread_init(NULL);
 
-  gethostname(hostname, MAXHOSTNAMELEN);
-  hent=gethostbyname(hostname);
-  if (!hent) {
-    g->thishost=owl_strdup("localhost");
-  } else {
-    g->thishost=owl_strdup(hent->h_name);
-  }
+  owl_select_init();
+
+  g->lines=LINES;
+  g->cols=COLS;
+  /* We shouldn't need this if we initialize lines and cols before the first
+   * owl_window_get_screen, but to be safe, we synchronize. */
+  owl_window_resize(owl_window_get_screen(), g->lines, g->cols);
 
   g->context_stack = NULL;
-  owl_global_push_context(g, OWL_CTX_STARTUP, NULL, NULL);
+  owl_global_push_context(g, OWL_CTX_STARTUP, NULL, NULL, NULL);
 
   g->curmsg=0;
   g->topmsg=0;
   g->markedmsgid=-1;
-  g->needrefresh=1;
   g->startupargs=NULL;
 
   owl_variable_dict_setup(&(g->vars));
 
-  g->lines=LINES;
-  g->cols=COLS;
-
   g->rightshift=0;
 
+  g->pw = NULL;
+  g->vw = NULL;
   g->tw = NULL;
 
   owl_keyhandler_init(&g->kh);
@@ -51,17 +40,12 @@ void owl_global_init(owl_global *g) {
 
   owl_dict_create(&(g->filters));
   g->filterlist = NULL;
-  owl_list_create(&(g->puntlist));
-  owl_list_create(&(g->messagequeue));
+  g->puntlist = g_ptr_array_new();
+  g->messagequeue = g_queue_new();
   owl_dict_create(&(g->styledict));
   g->curmsg_vert_offset=0;
   g->resizepending=0;
   g->direction=OWL_DIRECTION_DOWNWARDS;
-  g->zaway=0;
-  if (has_colors()) {
-    g->hascolors=1;
-  }
-  g->colorpairs=COLOR_PAIRS;
   owl_fmtext_init_colorpair_mgr(&(g->cpmgr));
   g->debug=OWL_DEBUG;
   owl_regex_init(&g->search_re);
@@ -70,28 +54,29 @@ void owl_global_init(owl_global *g) {
   g->newmsgproc_pid=0;
   
   owl_global_set_config_format(g, 0);
-  owl_global_set_userclue(g, OWL_USERCLUE_NONE);
   owl_global_set_no_have_config(g);
   owl_history_init(&(g->msghist));
   owl_history_init(&(g->cmdhist));
-  owl_history_set_norepeats(&(g->cmdhist));
   g->nextmsgid=0;
 
-  _owl_global_setup_windows(g);
-
   /* Fill in some variables which don't have constant defaults */
-  /* TODO: come back later and check passwd file first */
-  g->homedir=owl_strdup(getenv("HOME"));
+
+  /* glib's g_get_home_dir prefers passwd entries to $HOME, so we
+   * explicitly check getenv first. */
+  homedir = getenv("HOME");
+  if (!homedir)
+    homedir = g_get_home_dir();
+  g->homedir = g_strdup(homedir);
 
   g->confdir = NULL;
   g->startupfile = NULL;
-  cd = owl_sprintf("%s/%s", g->homedir, OWL_CONFIG_DIR);
+  cd = g_build_filename(g->homedir, OWL_CONFIG_DIR, NULL);
   owl_global_set_confdir(g, cd);
-  owl_free(cd);
+  g_free(cd);
 
-  owl_messagelist_create(&(g->msglist));
-  owl_mainwin_init(&(g->mw));
-  owl_popwin_init(&(g->pw));
+  g->msglist = owl_messagelist_new();
+
+  _owl_global_init_windows(g);
 
   g->aim_screenname=NULL;
   g->aim_screenname_for_filters=NULL;
@@ -104,20 +89,41 @@ void owl_global_init(owl_global *g) {
   owl_global_set_no_doaimevents(g);
 
   owl_errqueue_init(&(g->errqueue));
-  g->got_err_signal=0;
 
   owl_zbuddylist_create(&(g->zbuddies));
 
   g->zaldlist = NULL;
   g->pseudologin_notify = 0;
 
-  owl_obarray_init(&(g->obarray));
-
   owl_message_init_fmtext_cache();
-  owl_list_create(&(g->io_dispatch_list));
-  owl_list_create(&(g->psa_list));
-  g->timerlist = NULL;
-  g->interrupted = FALSE;
+  g->kill_buffer = NULL;
+
+  g->interrupt_count = 0;
+  g->interrupt_lock = g_mutex_new();
+}
+
+static void _owl_global_init_windows(owl_global *g)
+{
+  /* Create the main window */
+  owl_mainpanel_init(&(g->mainpanel));
+
+  /* Create the widgets */
+  g->mw = owl_mainwin_new(g->mainpanel.recwin);
+  owl_msgwin_init(&(g->msgwin), g->mainpanel.msgwin);
+  owl_sepbar_init(g->mainpanel.sepwin);
+
+  owl_window_set_default_cursor(g->mainpanel.sepwin);
+
+  /* set up a pad for input */
+  g->input_pad = newpad(1, 1);
+  nodelay(g->input_pad, 1);
+  keypad(g->input_pad, 1);
+  meta(g->input_pad, 1);
+}
+
+void owl_global_sepbar_dirty(owl_global *g)
+{
+  owl_window_dirty(g->mainpanel.sepwin);
 }
 
 /* Called once perl has been initialized */
@@ -126,104 +132,57 @@ void owl_global_complete_setup(owl_global *g)
   owl_cmddict_setup(&(g->cmds));
 }
 
-/* If *pan does not exist, we create a new panel, otherwise we replace the
-   window in *pan with win.
-
-   libpanel PANEL objects cannot exist without owner a valid window. This
-   maintains the invariant for _owl_global_setup_windows. */
-void _owl_panel_set_window(PANEL **pan, WINDOW *win)
-{
-  WINDOW *oldwin;
-
-  if (win == NULL) {
-    owl_function_debugmsg("_owl_panel_set_window: passed NULL win (failed to allocate?)\n");
-    endwin();
-    exit(50);
-  }
-
-  if (*pan) {
-    oldwin = panel_window(*pan);
-    replace_panel(*pan, win);
-    delwin(oldwin);
-  } else {
-    *pan = new_panel(win);
-  }
-}
-
-void _owl_global_setup_windows(owl_global *g) {
-  int cols, typwin_lines;
-
-  cols=g->cols;
-  typwin_lines=owl_global_get_typwin_lines(g);
-
-  /* set the new window sizes */
-  g->recwinlines=g->lines-(typwin_lines+2);
-  if (g->recwinlines<0) {
-    /* gotta deal with this */
-    g->recwinlines=0;
-  }
-
-  owl_function_debugmsg("_owl_global_setup_windows: about to call newwin(%i, %i, 0, 0)\n", g->recwinlines, cols);
-
-  /* create the new windows */
-  _owl_panel_set_window(&g->recpan, newwin(g->recwinlines, cols, 0, 0));
-  _owl_panel_set_window(&g->seppan, newwin(1, cols, g->recwinlines, 0));
-  _owl_panel_set_window(&g->msgpan, newwin(1, cols, g->recwinlines+1, 0));
-  _owl_panel_set_window(&g->typpan, newwin(typwin_lines, cols, g->recwinlines+2, 0));
-
-  if (g->tw)
-      owl_editwin_set_curswin(g->tw, owl_global_get_curs_typwin(g), typwin_lines, g->cols);
-
-  idlok(owl_global_get_curs_typwin(g), FALSE);
-  idlok(owl_global_get_curs_recwin(g), FALSE);
-  idlok(owl_global_get_curs_sepwin(g), FALSE);
-  idlok(owl_global_get_curs_msgwin(g), FALSE);
-
-  nodelay(owl_global_get_curs_typwin(g), 1);
-  keypad(owl_global_get_curs_typwin(g), TRUE);
-  wmove(owl_global_get_curs_typwin(g), 0, 0);
-
-  meta(owl_global_get_curs_typwin(g), TRUE);
-}
-
-owl_context *owl_global_get_context(owl_global *g) {
+owl_context *owl_global_get_context(const owl_global *g) {
   if (!g->context_stack)
     return NULL;
   return g->context_stack->data;
 }
 
-static void owl_global_lookup_keymap(owl_global *g) {
-  owl_context *c = owl_global_get_context(g);
-  if (!c || !c->keymap)
+static void owl_global_activate_context(owl_global *g, owl_context *c) {
+  if (!c)
     return;
 
-  if (!owl_keyhandler_activate(owl_global_get_keyhandler(g), c->keymap)) {
-    owl_function_error("Unable to activate keymap '%s'", c->keymap);
+  if (c->keymap) {
+    if (!owl_keyhandler_activate(owl_global_get_keyhandler(g), c->keymap)) {
+      owl_function_error("Unable to activate keymap '%s'", c->keymap);
+    }
   }
+  owl_window_set_cursor(c->cursor);
 }
 
-void owl_global_push_context(owl_global *g, int mode, void *data, const char *keymap) {
+void owl_global_push_context(owl_global *g, int mode, void *data, const char *keymap, owl_window *cursor) {
   owl_context *c;
-  if (!(mode & OWL_CTX_MODE_BITS))
-    mode |= OWL_CTX_INTERACTIVE;
-  c = owl_malloc(sizeof *c);
-  c->mode = mode;
-  c->data = data;
-  c->keymap = owl_strdup(keymap);
-  g->context_stack = g_list_prepend(g->context_stack, c);
-  owl_global_lookup_keymap(g);
+  c = owl_context_new(mode, data, keymap, cursor);
+  owl_global_push_context_obj(g, c);
 }
 
-void owl_global_pop_context(owl_global *g) {
+void owl_global_push_context_obj(owl_global *g, owl_context *c)
+{
+  g->context_stack = g_list_prepend(g->context_stack, c);
+  owl_global_activate_context(g, owl_global_get_context(g));
+}
+
+/* Pops the current context from the context stack and returns it. Caller is
+ * responsible for freeing. */
+CALLER_OWN owl_context *owl_global_pop_context_no_delete(owl_global *g)
+{
   owl_context *c;
   if (!g->context_stack)
-    return;
+    return NULL;
   c = owl_global_get_context(g);
+  owl_context_deactivate(c);
   g->context_stack = g_list_delete_link(g->context_stack,
                                         g->context_stack);
-  owl_free(c->keymap);
-  owl_free(c);
-  owl_global_lookup_keymap(g);
+  owl_global_activate_context(g, owl_global_get_context(g));
+  return c;
+}
+
+/* Pops the current context from the context stack and deletes it. */
+void owl_global_pop_context(owl_global *g) {
+  owl_context *c;
+  c = owl_global_pop_context_no_delete(g);
+  if (c)
+    owl_context_delete(c);
 }
 
 int owl_global_get_lines(const owl_global *g) {
@@ -235,7 +194,7 @@ int owl_global_get_cols(const owl_global *g) {
 }
 
 int owl_global_get_recwin_lines(const owl_global *g) {
-  return(g->recwinlines);
+  return g->mainpanel.recwinlines;
 }
 
 /* curmsg */
@@ -281,18 +240,23 @@ void owl_global_set_markedmsgid(owl_global *g, int i) {
 
 /* windows */
 
-owl_mainwin *owl_global_get_mainwin(owl_global *g) {
-  return(&(g->mw));
+owl_mainwin *owl_global_get_mainwin(owl_global *g)
+{
+  return g->mw;
 }
 
 owl_popwin *owl_global_get_popwin(owl_global *g) {
-  return(&(g->pw));
+  return g->pw;
+}
+
+void owl_global_set_popwin(owl_global *g, owl_popwin *pw) {
+  g->pw = pw;
 }
 
 /* msglist */
 
 owl_messagelist *owl_global_get_msglist(owl_global *g) {
-  return(&(g->msglist));
+  return g->msglist;
 }
 
 /* keyhandler */
@@ -301,43 +265,10 @@ owl_keyhandler *owl_global_get_keyhandler(owl_global *g) {
   return(&(g->kh));
 }
 
-/* curses windows */
-
-WINDOW *owl_global_get_curs_recwin(const owl_global *g) {
-  return panel_window(g->recpan);
-}
-
-WINDOW *owl_global_get_curs_sepwin(const owl_global *g) {
-  return panel_window(g->seppan);
-}
-
-WINDOW *owl_global_get_curs_msgwin(const owl_global *g) {
-  return panel_window(g->msgpan);
-}
-
-WINDOW *owl_global_get_curs_typwin(const owl_global *g) {
-  return panel_window(g->typpan);
-}
-
-/* typwin */
-
-owl_editwin *owl_global_get_typwin(const owl_global *g) {
-  return(g->tw);
-}
-
-/* refresh */
-
-int owl_global_is_needrefresh(const owl_global *g) {
-  if (g->needrefresh==1) return(1);
-  return(0);
-}
-
-void owl_global_set_needrefresh(owl_global *g) {
-  g->needrefresh=1;
-}
-
-void owl_global_set_noneedrefresh(owl_global *g) {
-  g->needrefresh=0;
+/* Gets the currently active typwin out of the current context. */
+owl_editwin *owl_global_current_typwin(const owl_global *g) {
+  owl_context *ctx = owl_global_get_context(g);
+  return owl_editcontext_get_editwin(ctx);
 }
 
 /* variable dictionary */
@@ -355,7 +286,8 @@ owl_cmddict *owl_global_get_cmddict(owl_global *g) {
 /* rightshift */
 
 void owl_global_set_rightshift(owl_global *g, int i) {
-  g->rightshift=i;
+  g->rightshift = i;
+  owl_mainwin_redisplay(owl_global_get_mainwin(g));
 }
 
 int owl_global_get_rightshift(const owl_global *g) {
@@ -370,7 +302,12 @@ owl_editwin *owl_global_set_typwin_active(owl_global *g, int style, owl_history
   if (d > 0 && style == OWL_EDITWIN_STYLE_MULTILINE)
       owl_function_resize_typwin(owl_global_get_typwin_lines(g) + d);
 
-  g->tw = owl_editwin_new(owl_global_get_curs_typwin(g),
+  if (g->typwin_erase_id) {
+    g_signal_handler_disconnect(g->mainpanel.typwin, g->typwin_erase_id);
+    g->typwin_erase_id = 0;
+  }
+
+  g->tw = owl_editwin_new(g->mainpanel.typwin,
                           owl_global_get_typwin_lines(g),
                           g->cols,
                           style,
@@ -378,19 +315,30 @@ owl_editwin *owl_global_set_typwin_active(owl_global *g, int style, owl_history
   return g->tw;
 }
 
+void owl_global_deactivate_editcontext(owl_context *ctx) {
+  owl_global *g = ctx->cbdata;
+  owl_global_set_typwin_inactive(g);
+}
+
 void owl_global_set_typwin_inactive(owl_global *g) {
   int d = owl_global_get_typewindelta(g);
   if (d > 0 && owl_editwin_get_style(g->tw) == OWL_EDITWIN_STYLE_MULTILINE)
       owl_function_resize_typwin(owl_global_get_typwin_lines(g) - d);
 
-  werase(owl_global_get_curs_typwin(g));
+  if (!g->typwin_erase_id) {
+    g->typwin_erase_id =
+      g_signal_connect(g->mainpanel.typwin, "redraw", G_CALLBACK(owl_window_erase_cb), NULL);
+  }
+  owl_window_dirty(g->mainpanel.typwin);
+
+  owl_editwin_unref(g->tw);
   g->tw = NULL;
 }
 
 /* resize */
 
 void owl_global_set_resize_pending(owl_global *g) {
-  g->resizepending=1;
+  g->resizepending = true;
 }
 
 const char *owl_global_get_homedir(const owl_global *g) {
@@ -407,10 +355,10 @@ const char *owl_global_get_confdir(const owl_global *g) {
  * Setting this also sets startupfile to confdir/startup
  */
 void owl_global_set_confdir(owl_global *g, const char *cd) {
-  owl_free(g->confdir);
-  g->confdir = owl_strdup(cd);
-  owl_free(g->startupfile);
-  g->startupfile = owl_sprintf("%s/startup", cd);
+  g_free(g->confdir);
+  g->confdir = g_strdup(cd);
+  g_free(g->startupfile);
+  g->startupfile = g_build_filename(cd, "startup", NULL);
 }
 
 const char *owl_global_get_startupfile(const owl_global *g) {
@@ -466,60 +414,36 @@ int owl_global_have_config(owl_global *g) {
   return(0);
 }
 
-void owl_global_resize(owl_global *g, int x, int y) {
-  /* resize the screen.  If x or y is 0 use the terminal size */
+/*
+ * Compute the size of the terminal. Try a ioctl, fallback to other stuff on
+ * fail.
+ */
+void owl_global_get_terminal_size(int *lines, int *cols) {
   struct winsize size;
-    
-  if (!g->resizepending) return;
-  g->resizepending = 0;
-
   /* get the new size */
   ioctl(STDIN_FILENO, TIOCGWINSZ, &size);
-  if (x==0) {
-    if (size.ws_row) {
-      g->lines=size.ws_row;
-    } else {
-      g->lines=LINES;
-    } 
+  if (size.ws_row) {
+    *lines = size.ws_row;
   } else {
-      g->lines=x;
+    *lines = LINES;
   }
 
-  if (y==0) {
-    if (size.ws_col) {
-      g->cols=size.ws_col;
-    } else {
-      g->cols=COLS;
-    } 
+  if (size.ws_col) {
+    *cols = size.ws_col;
   } else {
-    g->cols=y;
+    *cols = COLS;
   }
+}
 
-  resizeterm(size.ws_row, size.ws_col);
+void owl_global_check_resize(owl_global *g) {
+  /* resize the screen.  If lines or cols is 0 use the terminal size */
+  if (!g->resizepending) return;
+  g->resizepending = false;
 
-  /* re-initialize the windows */
-  _owl_global_setup_windows(g);
+  owl_global_get_terminal_size(&g->lines, &g->cols);
+  owl_window_resize(owl_window_get_screen(), g->lines, g->cols);
 
-  /* in case any styles rely on the current width */
-  owl_messagelist_invalidate_formats(owl_global_get_msglist(g));
-
-  /* recalculate the topmsg to make sure the current message is on
-   * screen */
-  owl_function_calculate_topmsg(OWL_DIRECTION_NONE);
-
-  /* refresh stuff */
-  g->needrefresh=1;
-  owl_mainwin_redisplay(&(g->mw));
-  sepbar(NULL);
-  if (g->tw)
-      owl_editwin_redisplay(g->tw);
-  else
-    werase(owl_global_get_curs_typwin(g));
-
-  owl_function_full_redisplay();
-
-  owl_function_debugmsg("New size is %i lines, %i cols.", size.ws_row, size.ws_col);
-  owl_function_makemsg("");
+  owl_function_debugmsg("New size is %i lines, %i cols.", g->lines, g->cols);
 }
 
 /* debug */
@@ -551,34 +475,14 @@ time_t owl_global_get_idletime(const owl_global *g) {
   return(time(NULL)-g->lastinputtime);
 }
 
-const char *owl_global_get_hostname(const owl_global *g) {
-  if (g->thishost) return(g->thishost);
-  return("");
-}
-
-/* userclue */
-
-void owl_global_set_userclue(owl_global *g, int clue) {
-  g->userclue=clue;
-}
-
-void owl_global_add_userclue(owl_global *g, int clue) {
-  g->userclue|=clue;
-}
-
-int owl_global_get_userclue(const owl_global *g) {
-  return(g->userclue);
-}
-
-int owl_global_is_userclue(const owl_global *g, int clue) {
-  if (g->userclue & clue) return(1);
-  return(0);
-}
-
 /* viewwin */
 
 owl_viewwin *owl_global_get_viewwin(owl_global *g) {
-  return(&(g->vw));
+  return g->vw;
+}
+
+void owl_global_set_viewwin(owl_global *g, owl_viewwin *vw) {
+  g->vw = vw;
 }
 
 
@@ -589,27 +493,14 @@ int owl_global_get_curmsg_vert_offset(const owl_global *g) {
 }
 
 void owl_global_set_curmsg_vert_offset(owl_global *g, int i) {
-  g->curmsg_vert_offset=i;
+  g->curmsg_vert_offset = i;
 }
 
 /* startup args */
 
-void owl_global_set_startupargs(owl_global *g, int argc, const char *const *argv) {
-  int i, len;
-
-  if (g->startupargs) owl_free(g->startupargs);
-  
-  len=0;
-  for (i=0; i<argc; i++) {
-    len+=strlen(argv[i])+5;
-  }
-  g->startupargs=owl_malloc(len+5);
-
-  strcpy(g->startupargs, "");
-  for (i=0; i<argc; i++) {
-    sprintf(g->startupargs + strlen(g->startupargs), "%s ", argv[i]);
-  }
-  g->startupargs[strlen(g->startupargs)-1]='\0';
+void owl_global_set_startupargs(owl_global *g, int argc, char **argv) {
+  g_free(g->startupargs);
+  g->startupargs = g_strjoinv(" ", argv);
 }
 
 const char *owl_global_get_startupargs(const owl_global *g) {
@@ -644,11 +535,11 @@ static void owl_global_delete_filter_ent(void *data)
   owl_global_filter_ent *e = data;
   e->g->filterlist = g_list_remove(e->g->filterlist, e->f);
   owl_filter_delete(e->f);
-  owl_free(e);
+  g_free(e);
 }
 
 void owl_global_add_filter(owl_global *g, owl_filter *f) {
-  owl_global_filter_ent *e = owl_malloc(sizeof *e);
+  owl_global_filter_ent *e = g_new(owl_global_filter_ent, 1);
   e->g = g;
   e->f = f;
 
@@ -675,39 +566,25 @@ owl_view *owl_global_get_current_view(owl_global *g) {
   return(&(g->current_view));
 }
 
-/* has colors */
-
-int owl_global_get_hascolors(const owl_global *g) {
-  if (g->hascolors) return(1);
-  return(0);
-}
-
-/* color pairs */
-
-int owl_global_get_colorpairs(const owl_global *g) {
-  return(g->colorpairs);
-}
-
 owl_colorpair_mgr *owl_global_get_colorpair_mgr(owl_global *g) {
   return(&(g->cpmgr));
 }
 
 /* puntlist */
 
-owl_list *owl_global_get_puntlist(owl_global *g) {
-  return(&(g->puntlist));
+GPtrArray *owl_global_get_puntlist(owl_global *g) {
+  return g->puntlist;
 }
 
 int owl_global_message_is_puntable(owl_global *g, const owl_message *m) {
-  const owl_list *pl;
-  int i, j;
+  const GPtrArray *pl;
+  int i;
 
-  pl=owl_global_get_puntlist(g);
-  j=owl_list_get_size(pl);
-  for (i=0; i<j; i++) {
-    if (owl_filter_message_match(owl_list_get_element(pl, i), m)) return(1);
+  pl = owl_global_get_puntlist(g);
+  for (i = 0; i < pl->len; i++) {
+    if (owl_filter_message_match(pl->pdata[i], m)) return 1;
   }
-  return(0);
+  return 0;
 }
 
 int owl_global_should_followlast(owl_global *g) {
@@ -733,6 +610,10 @@ void owl_global_set_search_re(owl_global *g, const owl_regex *re) {
   }
   if (re != NULL)
     owl_regex_copy(re, &g->search_re);
+  /* TODO: Emit a signal so we don't depend on the viewwin and mainwin */
+  if (owl_global_get_viewwin(g))
+    owl_viewwin_dirty(owl_global_get_viewwin(g));
+  owl_mainwin_redisplay(owl_global_get_mainwin(g));
 }
 
 const owl_regex *owl_global_get_search_re(const owl_global *g) {
@@ -774,15 +655,13 @@ const char *owl_global_get_aim_screenname_for_filters(const owl_global *g)
 void owl_global_set_aimloggedin(owl_global *g, const char *screenname)
 {
   char *sn_escaped;
-  const char *quote;
   g->aim_loggedin=1;
-  if (g->aim_screenname) owl_free(g->aim_screenname);
-  if (g->aim_screenname_for_filters) owl_free(g->aim_screenname_for_filters);
-  g->aim_screenname=owl_strdup(screenname);
+  if (g->aim_screenname) g_free(g->aim_screenname);
+  if (g->aim_screenname_for_filters) g_free(g->aim_screenname_for_filters);
+  g->aim_screenname=g_strdup(screenname);
   sn_escaped = owl_text_quote(screenname, OWL_REGEX_QUOTECHARS, OWL_REGEX_QUOTEWITH);
-  quote = owl_getquoting(sn_escaped);
-  g->aim_screenname_for_filters=owl_sprintf("%s%s%s", quote, sn_escaped, quote);
-  owl_free(sn_escaped);
+  g->aim_screenname_for_filters = owl_arg_quote(sn_escaped);
+  g_free(sn_escaped);
 }
 
 void owl_global_set_aimnologgedin(owl_global *g)
@@ -790,20 +669,26 @@ void owl_global_set_aimnologgedin(owl_global *g)
   g->aim_loggedin=0;
 }
 
-int owl_global_is_doaimevents(const owl_global *g)
+bool owl_global_is_doaimevents(const owl_global *g)
 {
-  if (g->aim_doprocessing) return(1);
-  return(0);
+  return g->aim_event_source != NULL;
 }
 
 void owl_global_set_doaimevents(owl_global *g)
 {
-  g->aim_doprocessing=1;
+  if (g->aim_event_source)
+    return;
+  g->aim_event_source = owl_aim_event_source_new(owl_global_get_aimsess(g));
+  g_source_attach(g->aim_event_source, NULL);
 }
 
 void owl_global_set_no_doaimevents(owl_global *g)
 {
-  g->aim_doprocessing=0;
+  if (!g->aim_event_source)
+    return;
+  g_source_destroy(g->aim_event_source);
+  g_source_unref(g->aim_event_source);
+  g->aim_event_source = NULL;
 }
 
 aim_session_t *owl_global_get_aimsess(owl_global *g)
@@ -825,27 +710,26 @@ void owl_global_set_bossconn(owl_global *g, aim_conn_t *conn)
 
 void owl_global_messagequeue_addmsg(owl_global *g, owl_message *m)
 {
-  owl_list_append_element(&(g->messagequeue), m);
+  g_queue_push_tail(g->messagequeue, m);
 }
 
 /* pop off the first message and return it.  Return NULL if the queue
  * is empty.  The caller should free the message after using it, if
  * necessary.
  */
-owl_message *owl_global_messagequeue_popmsg(owl_global *g)
+CALLER_OWN owl_message *owl_global_messagequeue_popmsg(owl_global *g)
 {
   owl_message *out;
 
-  if (owl_list_get_size(&(g->messagequeue))==0) return(NULL);
-  out=owl_list_get_element(&(g->messagequeue), 0);
-  owl_list_remove_element(&(g->messagequeue), 0);
-  return(out);
+  if (g_queue_is_empty(g->messagequeue))
+    return NULL;
+  out = g_queue_pop_head(g->messagequeue);
+  return out;
 }
 
 int owl_global_messagequeue_pending(owl_global *g)
 {
-  if (owl_list_get_size(&(g->messagequeue))==0) return(0);
-  return(1);
+  return !g_queue_is_empty(g->messagequeue);
 }
 
 owl_buddylist *owl_global_get_buddylist(owl_global *g)
@@ -862,10 +746,9 @@ const owl_style *owl_global_get_style_by_name(const owl_global *g, const char *n
   return owl_dict_find_element(&(g->styledict), name);
 }
 
-/* creates a list and fills it in with keys.  duplicates the keys, 
- * so they will need to be freed by the caller. */
-int owl_global_get_style_names(const owl_global *g, owl_list *l) {
-  return owl_dict_get_keys(&(g->styledict), l);
+CALLER_OWN GPtrArray *owl_global_get_style_names(const owl_global *g)
+{
+  return owl_dict_get_keys(&g->styledict);
 }
 
 void owl_global_add_style(owl_global *g, owl_style *s)
@@ -924,28 +807,6 @@ owl_errqueue *owl_global_get_errqueue(owl_global *g)
   return(&(g->errqueue));
 }
 
-void owl_global_set_errsignal(owl_global *g, int signum, siginfo_t *siginfo)
-{
-  g->got_err_signal = signum;
-  if (siginfo) {
-    g->err_signal_info = *siginfo;
-  } else {
-    memset(&(g->err_signal_info), 0, sizeof(siginfo_t));
-  }
-}
-
-int owl_global_get_errsignal_and_clear(owl_global *g, siginfo_t *siginfo)
-{
-  int signum;
-  if (siginfo && g->got_err_signal) {
-    *siginfo = g->err_signal_info;
-  } 
-  signum = g->got_err_signal;
-  g->got_err_signal = 0;
-  return signum;
-}
-
-
 owl_zbuddylist *owl_global_get_zephyr_buddylist(owl_global *g)
 {
   return(&(g->zbuddies));
@@ -971,36 +832,35 @@ struct termios *owl_global_get_startup_tio(owl_global *g)
   return(&(g->startup_tio));
 }
 
-const char * owl_global_intern(owl_global *g, const char * string)
+void owl_global_setup_default_filters(owl_global *g)
 {
-  return owl_obarray_insert(&(g->obarray), string);
-}
+  int i;
+  static const struct {
+    const char *name;
+    const char *desc;
+  } filters[] = {
+    { "personal",
+      "isprivate ^true$ and ( not type ^zephyr$ or ( class ^message  ) )" },
+    { "trash",
+      "class ^mail$ or opcode ^ping$ or type ^admin$ or ( not login ^none$ )" },
+    { "wordwrap", "not ( type ^admin$ or type ^zephyr$ )" },
+    { "ping", "opcode ^ping$" },
+    { "auto", "opcode ^auto$" },
+    { "login", "not login ^none$" },
+    { "reply-lockout", "class ^mail$" },
+    { "out", "direction ^out$" },
+    { "aim", "type ^aim$" },
+    { "zephyr", "type ^zephyr$" },
+    { "none", "false" },
+    { "all", "true" },
+    { NULL, NULL }
+  };
 
-owl_list *owl_global_get_io_dispatch_list(owl_global *g)
-{
-  return &(g->io_dispatch_list);
-}
+  owl_function_debugmsg("startup: creating default filters");
 
-owl_list *owl_global_get_psa_list(owl_global *g)
-{
-  return &(g->psa_list);
-}
-
-GList **owl_global_get_timerlist(owl_global *g)
-{
-  return &(g->timerlist);
-}
-
-int owl_global_is_interrupted(const owl_global *g) {
-  return g->interrupted;
-}
-
-void owl_global_set_interrupted(owl_global *g) {
-  g->interrupted = 1;
-}
-
-void owl_global_unset_interrupted(owl_global *g) {
-  g->interrupted = 0;
+  for (i = 0; filters[i].name != NULL; i++)
+    owl_global_add_filter(g, owl_filter_new_fromstring(filters[i].name,
+                                                       filters[i].desc));
 }
 
 FILE *owl_global_get_debug_file_handle(owl_global *g) {
@@ -1016,15 +876,43 @@ FILE *owl_global_get_debug_file_handle(owl_global *g) {
 
     g->debug_file = NULL;
 
-    path = owl_sprintf("%s.%d", filename, getpid());
+    path = g_strdup_printf("%s.%d", filename, getpid());
     fd = open(path, O_CREAT|O_WRONLY|O_EXCL, 0600);
-    owl_free(path);
+    g_free(path);
 
     if (fd >= 0)
       g->debug_file = fdopen(fd, "a");
 
-    owl_free(open_file);
-    open_file = owl_strdup(filename);
+    g_free(open_file);
+    open_file = g_strdup(filename);
   }
   return g->debug_file;
 }
+
+const char *owl_global_get_kill_buffer(owl_global *g) {
+  return g->kill_buffer;
+}
+
+void owl_global_set_kill_buffer(owl_global *g, const char *kill, int len) {
+  g_free(g->kill_buffer);
+  g->kill_buffer = g_strndup(kill, len);
+}
+
+void owl_global_add_interrupt(owl_global *g) {
+  /* TODO: This can almost certainly be done with atomic
+   * operations. Whatever. */
+  g_mutex_lock(g->interrupt_lock);
+  g->interrupt_count++;
+  g_mutex_unlock(g->interrupt_lock);
+}
+
+bool owl_global_take_interrupt(owl_global *g) {
+  bool ans = false;
+  g_mutex_lock(g->interrupt_lock);
+  if (g->interrupt_count > 0) {
+    ans = true;
+    g->interrupt_count--;
+  }
+  g_mutex_unlock(g->interrupt_lock);
+  return ans;
+}
