@@ -1,15 +1,7 @@
-#include <stdlib.h>
-#include <unistd.h>
-#include <string.h>
-#include <sys/socket.h>
-#include <netdb.h>
-#include <sys/types.h>
-#include <sys/socket.h>
-#include <netinet/in.h>
-#include <arpa/inet.h>
-#include <time.h>
 #include "owl.h"
 #include "filterproc.h"
+#include <sys/socket.h>
+#include <arpa/inet.h>
 
 static owl_fmtext_cache fmtext_cache[OWL_FMTEXT_CACHE_SIZE];
 static owl_fmtext_cache * fmtext_cache_next = fmtext_cache;
@@ -41,13 +33,17 @@ void owl_message_init(owl_message *m)
   owl_message_set_direction_none(m);
   m->delete=0;
 
+#ifdef HAVE_LIBZEPHYR
+  m->has_notice = false;
+#endif
+
   owl_message_set_hostname(m, "");
-  owl_list_create(&(m->attributes));
+  m->attributes = g_ptr_array_new();
   
   /* save the time */
-  m->time=time(NULL);
-  m->timestr=owl_strdup(ctime(&(m->time)));
-  m->timestr[strlen(m->timestr)-1]='\0';
+  m->time = time(NULL);
+  m->timestr = g_strdup(ctime(&m->time));
+  m->timestr[strlen(m->timestr)-1] = '\0';
 
   m->fmtext = NULL;
 }
@@ -57,24 +53,25 @@ void owl_message_init(owl_message *m)
  */
 void owl_message_set_attribute(owl_message *m, const char *attrname, const char *attrvalue)
 {
-  int i, j;
+  int i;
   owl_pair *p = NULL, *pair = NULL;
 
+  attrname = g_intern_string(attrname);
+
   /* look for an existing pair with this key, */
-  j=owl_list_get_size(&(m->attributes));
-  for (i=0; i<j; i++) {
-    p=owl_list_get_element(&(m->attributes), i);
-    if (!strcmp(owl_pair_get_key(p), attrname)) {
-      owl_free(owl_pair_get_value(p));
+  for (i = 0; i < m->attributes->len; i++) {
+    p = m->attributes->pdata[i];
+    if (owl_pair_get_key(p) == attrname) {
+      g_free(owl_pair_get_value(p));
       pair = p;
       break;
     }
   }
 
   if(pair ==  NULL) {
-    pair = owl_malloc(sizeof(owl_pair));
-    owl_pair_create(pair, owl_global_intern(&g, attrname), NULL);
-    owl_list_append_element(&(m->attributes), pair);
+    pair = g_new(owl_pair, 1);
+    owl_pair_create(pair, attrname, NULL);
+    g_ptr_array_add(m->attributes, pair);
   }
   owl_pair_set_value(pair, owl_validate_or_convert(attrvalue));
 }
@@ -84,13 +81,19 @@ void owl_message_set_attribute(owl_message *m, const char *attrname, const char
  */
 const char *owl_message_get_attribute_value(const owl_message *m, const char *attrname)
 {
-  int i, j;
+  int i;
   owl_pair *p;
+  GQuark quark;
 
-  j=owl_list_get_size(&(m->attributes));
-  for (i=0; i<j; i++) {
-    p=owl_list_get_element(&(m->attributes), i);
-    if (!strcmp(owl_pair_get_key(p), attrname)) {
+  quark = g_quark_try_string(attrname);
+  if (quark == 0)
+    /* don't bother inserting into string table */
+    return NULL;
+  attrname = g_quark_to_string(quark);
+
+  for (i = 0; i < m->attributes->len; i++) {
+    p = m->attributes->pdata[i];
+    if (owl_pair_get_key(p) == attrname) {
       return(owl_pair_get_value(p));
     }
   }
@@ -108,23 +111,30 @@ const char *owl_message_get_attribute_value(const owl_message *m, const char *at
  * function to indent fmtext.
  */
 void owl_message_attributes_tofmtext(const owl_message *m, owl_fmtext *fm) {
-  int i, j;
+  int i;
   owl_pair *p;
-  char *buff;
+  char *buff, *tmpbuff;
 
   owl_fmtext_init_null(fm);
 
-  j=owl_list_get_size(&(m->attributes));
-  for (i=0; i<j; i++) {
-    p=owl_list_get_element(&(m->attributes), i);
-    buff=owl_sprintf("  %-15.15s: %-35.35s\n", owl_pair_get_key(p), owl_pair_get_value(p));
+  for (i = 0; i < m->attributes->len; i++) {
+    p = m->attributes->pdata[i];
+
+    buff = g_strdup(owl_pair_get_value(p));
+    if (buff) {
+      tmpbuff = owl_text_indent(buff, 19, false);
+      g_free(buff);
+      buff = g_strdup_printf("  %-15.15s: %s\n", owl_pair_get_key(p), tmpbuff);
+      g_free(tmpbuff);
+    }
+
     if(buff == NULL) {
-      buff=owl_sprintf("  %-15.15s: %-35.35s\n", owl_pair_get_key(p), "<error>");
+      buff = g_strdup_printf("  %-15.15s: %s\n", owl_pair_get_key(p), "<error>");
       if(buff == NULL)
-        buff=owl_strdup("   <error>\n");
+        buff=g_strdup("   <error>\n");
     }
     owl_fmtext_append_normal(fm, buff);
-    owl_free(buff);
+    g_free(buff);
   }
 }
 
@@ -336,6 +346,11 @@ const char *owl_message_get_timestr(const owl_message *m)
   return("");
 }
 
+CALLER_OWN char *owl_message_format_time(const owl_message *m)
+{
+  return owl_util_format_time(localtime(&m->time));
+}
+
 void owl_message_set_type_admin(owl_message *m)
 {
   owl_message_set_attribute(m, "type", "admin");
@@ -484,7 +499,7 @@ int owl_message_is_delete(const owl_message *m)
 #ifdef HAVE_LIBZEPHYR
 const ZNotice_t *owl_message_get_notice(const owl_message *m)
 {
-  return(&(m->notice));
+  return m->has_notice ? &m->notice : NULL;
 }
 #else
 void *owl_message_get_notice(const owl_message *m)
@@ -495,7 +510,7 @@ void *owl_message_get_notice(const owl_message *m)
 
 void owl_message_set_hostname(owl_message *m, const char *hostname)
 {
-  m->hostname=owl_global_intern(&g, hostname);
+  m->hostname = g_intern_string(hostname);
 }
 
 const char *owl_message_get_hostname(const owl_message *m)
@@ -515,10 +530,8 @@ void owl_message_curs_waddstr(owl_message *m, WINDOW *win, int aline, int bline,
   
   owl_fmtext_truncate_lines(&(m->fmtext->fmtext), aline, bline-aline, &a);
   owl_fmtext_truncate_cols(&a, acol, bcol, &b);
-  owl_fmtext_colorize(&b, fgcolor);
-  owl_fmtext_colorizebg(&b, bgcolor);
 
-  owl_fmtext_curs_waddstr(&b, win);
+  owl_fmtext_curs_waddstr(&b, win, OWL_FMTEXT_ATTR_NONE, fgcolor, bgcolor);
 
   owl_fmtext_cleanup(&a);
   owl_fmtext_cleanup(&b);
@@ -566,7 +579,7 @@ int owl_message_is_mail(const owl_message *m)
 }
 
 /* caller must free return value. */
-char *owl_message_get_cc(const owl_message *m)
+CALLER_OWN char *owl_message_get_cc(const owl_message *m)
 {
   const char *cur;
   char *out, *end;
@@ -576,46 +589,37 @@ char *owl_message_get_cc(const owl_message *m)
   if (strncasecmp(cur, "cc:", 3)) return(NULL);
   cur+=3;
   while (*cur && *cur==' ') cur++;
-  out = owl_strdup(cur);
+  out = g_strdup(cur);
   end = strchr(out, '\n');
   if (end) end[0] = '\0';
   return(out);
 }
 
 /* caller must free return value */
-char *owl_message_get_cc_without_recipient(const owl_message *m)
+CALLER_OWN GList *owl_message_get_cc_without_recipient(const owl_message *m)
 {
-  char *cc, *out, *end, *shortuser, *recip;
+  char *cc, *shortuser, *recip;
   const char *user;
+  GList *out = NULL;
 
   cc = owl_message_get_cc(m);
   if (cc == NULL)
     return NULL;
 
   recip = short_zuser(owl_message_get_recipient(m));
-  out = owl_malloc(strlen(cc) + 2);
-  end = out;
 
   user = strtok(cc, " ");
   while (user != NULL) {
     shortuser = short_zuser(user);
     if (strcasecmp(shortuser, recip) != 0) {
-      strcpy(end, user);
-      end[strlen(user)] = ' ';
-      end += strlen(user) + 1;
+      out = g_list_prepend(out, g_strdup(user));
     }
-    owl_free(shortuser);
+    g_free(shortuser);
     user = strtok(NULL, " ");
   }
-  end[0] = '\0';
 
-  owl_free(recip);
-  owl_free(cc);
-
-  if (strlen(out) == 0) {
-    owl_free(out);
-    out = NULL;
-  }
+  g_free(recip);
+  g_free(cc);
 
   return(out);
 }
@@ -680,7 +684,7 @@ int owl_message_search(owl_message *m, const owl_regex *re)
 
   owl_message_format(m); /* is this necessary? */
   
-  return (owl_fmtext_search(&(m->fmtext->fmtext), re));
+  return owl_fmtext_search(&(m->fmtext->fmtext), re, 0) >= 0;
 }
 
 
@@ -733,6 +737,41 @@ void owl_message_create_loopback(owl_message *m, const char *text)
   owl_message_set_isprivate(m);
 }
 
+void owl_message_save_ccs(owl_message *m) {
+  GList *cc;
+  char *tmp;
+
+  cc = owl_message_get_cc_without_recipient(m);
+
+  if (cc != NULL) {
+    GString *recips = g_string_new("");
+    cc = g_list_prepend(cc, short_zuser(owl_message_get_sender(m)));
+    cc = g_list_prepend(cc, short_zuser(owl_message_get_recipient(m)));
+    cc = g_list_sort(cc, (GCompareFunc)strcasecmp);
+
+    while(cc != NULL) {
+      /* Collapse any identical entries */
+      while (cc->next && strcasecmp(cc->data, cc->next->data) == 0) {
+        g_free(cc->data);
+        cc = g_list_delete_link(cc, cc);
+      }
+
+      tmp = short_zuser(cc->data);
+      g_string_append(recips, tmp);
+
+      g_free(tmp);
+      g_free(cc->data);
+      cc = g_list_delete_link(cc, cc);
+
+      if (cc)
+        g_string_append_c(recips, ' ');
+    }
+
+    owl_message_set_attribute(m, "zephyr_ccs", recips->str);
+    g_string_free(recips, true);
+  }
+}
+
 #ifdef HAVE_LIBZEPHYR
 void owl_message_create_from_znotice(owl_message *m, const ZNotice_t *n)
 {
@@ -741,7 +780,6 @@ void owl_message_create_from_znotice(owl_message *m, const ZNotice_t *n)
 #else /* !ZNOTICE_SOCKADDR */
   struct hostent *hent;
 #endif /* ZNOTICE_SOCKADDR */
-  const char *ptr;
   char *tmp, *tmp2;
   int len;
 
@@ -752,15 +790,16 @@ void owl_message_create_from_znotice(owl_message *m, const ZNotice_t *n)
   
   /* first save the full notice */
   m->notice = *n;
+  m->has_notice = true;
 
   /* a little gross, we'll replace \r's with ' ' for now */
   owl_zephyr_hackaway_cr(&(m->notice));
   
   /* save the time, we need to nuke the string saved by message_init */
-  if (m->timestr) owl_free(m->timestr);
-  m->time=n->z_time.tv_sec;
-  m->timestr=owl_strdup(ctime(&(m->time)));
-  m->timestr[strlen(m->timestr)-1]='\0';
+  if (m->timestr) g_free(m->timestr);
+  m->time = n->z_time.tv_sec;
+  m->timestr = g_strdup(ctime(&m->time));
+  m->timestr[strlen(m->timestr)-1] = '\0';
 
   /* set other info */
   owl_message_set_sender(m, n->z_sender);
@@ -774,22 +813,18 @@ void owl_message_create_from_znotice(owl_message *m, const ZNotice_t *n)
   }
   owl_message_set_zsig(m, owl_zephyr_get_zsig(n, &len));
 
-  if ((ptr=strchr(n->z_recipient, '@'))!=NULL) {
-    owl_message_set_realm(m, ptr+1);
-  } else {
-    owl_message_set_realm(m, owl_zephyr_get_realm());
-  }
+  owl_message_set_realm(m, zuser_realm(n->z_recipient));
 
   /* Set the "isloginout" attribute if it's a login message */
   if (!strcasecmp(n->z_class, "login") || !strcasecmp(n->z_class, OWL_WEBZEPHYR_CLASS)) {
     if (!strcasecmp(n->z_opcode, "user_login") || !strcasecmp(n->z_opcode, "user_logout")) {
       tmp=owl_zephyr_get_field(n, 1);
       owl_message_set_attribute(m, "loginhost", tmp);
-      owl_free(tmp);
+      g_free(tmp);
 
       tmp=owl_zephyr_get_field(n, 3);
       owl_message_set_attribute(m, "logintty", tmp);
-      owl_free(tmp);
+      g_free(tmp);
     }
 
     if (!strcasecmp(n->z_opcode, "user_login")) {
@@ -800,10 +835,8 @@ void owl_message_create_from_znotice(owl_message *m, const ZNotice_t *n)
   }
 
   
-  /* set the "isprivate" attribute if it's a private zephyr.
-   ``private'' means recipient is non-empty and doesn't start wit 
-   `@' */
-  if (*n->z_recipient && *n->z_recipient != '@') {
+  /* set the "isprivate" attribute if it's a private zephyr. */
+  if (owl_zwrite_recip_is_personal(n->z_recipient)) {
     owl_message_set_isprivate(m);
   }
 
@@ -832,30 +865,31 @@ void owl_message_create_from_znotice(owl_message *m, const ZNotice_t *n)
   if (owl_global_is_newlinestrip(&g)) {
     tmp2=owl_util_stripnewlines(tmp);
     owl_message_set_body(m, tmp2);
-    owl_free(tmp2);
+    g_free(tmp2);
   } else {
     owl_message_set_body(m, tmp);
   }
-  owl_free(tmp);
+  g_free(tmp);
 
   /* if zcrypt is enabled try to decrypt the message */
   if (owl_global_is_zcrypt(&g) && !strcasecmp(n->z_opcode, "crypt")) {
     const char *argv[] = {
-      "zcrypt",
+      NULL,
       "-D",
       "-c", owl_message_get_class(m),
       "-i", owl_message_get_instance(m),
       NULL
     };
-    char *out;
+    char *out = NULL;
     int rv;
     int status;
     char *zcrypt;
 
-    zcrypt = owl_sprintf("%s/zcrypt", owl_get_bindir());
+    zcrypt = g_build_filename(owl_get_bindir(), "zcrypt", NULL);
+    argv[0] = zcrypt;
 
-    rv = call_filter(zcrypt, argv, owl_message_get_body(m), &out, &status);
-    owl_free(zcrypt);
+    rv = call_filter(argv, owl_message_get_body(m), &out, &status);
+    g_free(zcrypt);
 
     if(!rv && !status) {
       int len = strlen(out);
@@ -863,11 +897,15 @@ void owl_message_create_from_znotice(owl_message *m, const ZNotice_t *n)
         out[len - 8] = 0;
       }
       owl_message_set_body(m, out);
-      owl_free(out);
-    } else if(out) {
-      owl_free(out);
+    } else {
+      /* Replace the opcode. Otherwise the UI and other bits of code think the
+       * message was encrypted. */
+      owl_message_set_opcode(m, "failed-decrypt");
     }
+    g_free(out);
   }
+
+  owl_message_save_ccs(m);
 }
 #else
 void owl_message_create_from_znotice(owl_message *m, const void *n)
@@ -879,7 +917,6 @@ void owl_message_create_from_znotice(owl_message *m, const void *n)
 void owl_message_create_pseudo_zlogin(owl_message *m, int direction, const char *user, const char *host, const char *time, const char *tty)
 {
   char *longuser;
-  const char *ptr;
 
 #ifdef HAVE_LIBZEPHYR
   memset(&(m->notice), 0, sizeof(ZNotice_t));
@@ -908,24 +945,19 @@ void owl_message_create_pseudo_zlogin(owl_message *m, int direction, const char
     owl_message_set_islogout(m);
   }
 
-  if ((ptr=strchr(longuser, '@'))!=NULL) {
-    owl_message_set_realm(m, ptr+1);
-  } else {
-    owl_message_set_realm(m, owl_zephyr_get_realm());
-  }
+  owl_message_set_realm(m, zuser_realm(longuser));
 
   owl_message_set_body(m, "<uninitialized>");
 
   /* save the hostname */
   owl_function_debugmsg("create_pseudo_login: host is %s", host ? host : "");
   owl_message_set_hostname(m, host ? host : "");
-  owl_free(longuser);
+  g_free(longuser);
 }
 
-void owl_message_create_from_zwrite(owl_message *m, const owl_zwrite *z, const char *body)
+void owl_message_create_from_zwrite(owl_message *m, const owl_zwrite *z, const char *body, int recip_index)
 {
-  int ret;
-  char hostbuff[5000];
+  char *replyline;
   
   owl_message_init(m);
 
@@ -935,55 +967,61 @@ void owl_message_create_from_zwrite(owl_message *m, const owl_zwrite *z, const c
   owl_message_set_sender(m, owl_zephyr_get_sender());
   owl_message_set_class(m, owl_zwrite_get_class(z));
   owl_message_set_instance(m, owl_zwrite_get_instance(z));
-  if (owl_zwrite_get_numrecips(z)>0) {
-    char *longzuser = long_zuser(owl_zwrite_get_recip_n(z, 0));
-    owl_message_set_recipient(m,
-			      longzuser); /* only gets the first user, must fix */
-    owl_free(longzuser);
+  if (recip_index < owl_zwrite_get_numrecips(z)) {
+    char *zuser = owl_zwrite_get_recip_n_with_realm(z, recip_index);
+    char *longzuser = long_zuser(zuser);
+    owl_message_set_recipient(m, longzuser);
+    owl_message_set_realm(m, zuser_realm(longzuser));
+    g_free(longzuser);
+    g_free(zuser);
+  } else {
+    /* TODO: We should maybe munge this into the case above, as that comparison
+     * is a little overly clever. It's also not clear this codepath ever runs
+     * anyway. */
+    const char *realm = owl_zwrite_get_realm(z);
+    owl_message_set_realm(m, realm[0] ? realm : owl_zephyr_get_realm());
   }
   owl_message_set_opcode(m, owl_zwrite_get_opcode(z));
-  owl_message_set_realm(m, owl_zwrite_get_realm(z)); /* also a hack, but not here */
-  if(z->zwriteline) {
-    owl_message_set_zwriteline(m, z->zwriteline);
-  }
+
+  /* Although not strictly the zwriteline, anyone using the unsantized version
+   * of it probably has a bug. */
+  replyline = owl_zwrite_get_replyline(z, recip_index);
+  owl_message_set_zwriteline(m, replyline);
+  g_free(replyline);
+
   owl_message_set_body(m, body);
   owl_message_set_zsig(m, owl_zwrite_get_zsig(z));
   
   /* save the hostname */
-  ret=gethostname(hostbuff, MAXHOSTNAMELEN);
-  hostbuff[MAXHOSTNAMELEN]='\0';
-  if (ret) {
-    owl_message_set_hostname(m, "localhost");
-  } else {
-    owl_message_set_hostname(m, hostbuff);
-  }
+  owl_message_set_hostname(m, g_get_host_name());
 
   /* set the "isprivate" attribute if it's a private zephyr. */
   if (owl_zwrite_is_personal(z)) {
     owl_message_set_isprivate(m);
   }
+
+  owl_message_save_ccs(m);
 }
 
 void owl_message_cleanup(owl_message *m)
 {
-  int i, j;
+  int i;
   owl_pair *p;
 #ifdef HAVE_LIBZEPHYR    
-  if (owl_message_is_type_zephyr(m) && owl_message_is_direction_in(m)) {
+  if (m->has_notice) {
     ZFreeNotice(&(m->notice));
   }
 #endif
-  if (m->timestr) owl_free(m->timestr);
+  if (m->timestr) g_free(m->timestr);
 
   /* free all the attributes */
-  j=owl_list_get_size(&(m->attributes));
-  for (i=0; i<j; i++) {
-    p=owl_list_get_element(&(m->attributes), i);
-    owl_free(owl_pair_get_value(p));
-    owl_free(p);
+  for (i = 0; i < m->attributes->len; i++) {
+    p = m->attributes->pdata[i];
+    g_free(owl_pair_get_value(p));
+    g_free(p);
   }
 
-  owl_list_cleanup(&(m->attributes), NULL);
+  g_ptr_array_free(m->attributes, true);
  
   owl_message_invalidate_format(m);
 }
@@ -991,5 +1029,5 @@ void owl_message_cleanup(owl_message *m)
 void owl_message_delete(owl_message *m)
 {
   owl_message_cleanup(m);
-  owl_free(m);
+  g_free(m);
 }
