@@ -4,10 +4,11 @@ use warnings;
 package BarnOwl;
 
 use base qw(Exporter);
-our @EXPORT_OK = qw(command getcurmsg getnumcols getidletime
+our @EXPORT_OK = qw(command getcurmsg getnumcols getnumlines getidletime
                     zephyr_getsender zephyr_getrealm zephyr_zwrite
                     zephyr_stylestrip zephyr_smartstrip_user zephyr_getsubs
                     queue_message admin_message
+                    start_edit
                     start_question start_password start_edit_win
                     get_data_dir get_config_dir popless_text popless_ztext
                     error debug
@@ -28,6 +29,9 @@ BEGIN {
 use lib(get_data_dir() . "/lib");
 use lib(get_config_dir() . "/lib");
 
+use Glib;
+use AnyEvent;
+
 use BarnOwl::Hook;
 use BarnOwl::Hooks;
 use BarnOwl::Message;
@@ -48,7 +52,7 @@ BarnOwl
 
 The BarnOwl module contains the core of BarnOwl's perl
 bindings. Source in this module is also run at startup to bootstrap
-barnowl by defining things like the default style.
+BarnOwl by defining things like the default style.
 
 =for NOTE
 These following functions are defined in perlglue.xs. Keep the
@@ -76,11 +80,11 @@ seconds.
 
 =head2 zephyr_getrealm
 
-Returns the zephyr realm barnowl is running in
+Returns the zephyr realm BarnOwl is running in
 
 =head2 zephyr_getsender
 
-Returns the fully-qualified name of the zephyr sender barnowl is
+Returns the fully-qualified name of the zephyr sender BarnOwl is
 running as, e.g. C<nelhage@ATHENA.MIT.EDU>
 
 =head2 zephyr_zwrite COMMAND MESSAGE
@@ -88,6 +92,14 @@ running as, e.g. C<nelhage@ATHENA.MIT.EDU>
 Sends a zephyr programmatically. C<COMMAND> should be a C<zwrite>
 command line, and C<MESSAGE> is the zephyr body to send.
 
+=cut
+
+sub zephyr_zwrite {
+    my ($command, $message) = @_;
+    my $ret = BarnOwl::Internal::zephyr_zwrite($command, $message);
+    die "Error sending zephyr" unless $ret == 0;
+}
+
 =head2 ztext_stylestrip STRING
 
 Strips zephyr formatting from a string and returns the result
@@ -107,22 +119,92 @@ BarnOwl::Message or a subclass.
 
 Display a BarnOwl B<Admin> message, with the given header and body.
 
-=head2 start_question PROMPT CALLBACK
+=head2 start_edit %ARGS
 
-Displays C<PROMPT> on the screen and lets the user enter a line of
-text, and calls C<CALLBACK>, which must be a perl subroutine
-reference, with the text the user entered
+Displays a prompt on the screen and lets the user enter text,
+and calls a callback when the editwin is closed.
+
+C<%ARGS> must contain the following keys:
+
+=over 4
+
+=item prompt
+
+The line to display on the screen
+
+=item type
+
+One of:
+
+=over 4
+
+=item edit_win
+
+Displays the prompt on a line of its own and opens the edit_win.
+
+=item question
+
+Displays prompt on the screen and lets the user enter a line of
+text.
+
+=item password
+
+Like question, but echoes the user's input as C<*>s when they
+input.
+
+=back
+
+=item callback
+
+A Perl subroutine that is called when the user closes the edit_win.
+C<CALLBACK> gets called with two parameters: the text the user entered,
+and a C<SUCCESS> boolean parameter which is false if the user canceled
+the edit_win and true otherwise.
+
+=back
+
+=head2 start_question PROMPT CALLBACK
 
 =head2 start_password PROMPT CALLBACK
 
-Like C<start_question>, but echoes the user's input as C<*>s when they
-input.
-
 =head2 start_edit_win PROMPT CALLBACK
 
-Like C<start_question>, but displays C<PROMPT> on a line of its own
-and opens the editwin. If the user cancels the edit win, C<CALLBACK>
-is not invoked.
+Roughly equivalent to C<start_edit> called with the appropriate parameters.
+C<CALLBACK> is only called on success, for compatibility.
+
+These are deprecated wrappers around L<BarnOwl::start_edit>, and should not
+be uesd in new code.
+
+=cut
+
+sub start_edit {
+    my %args = (@_);
+    BarnOwl::Internal::start_edit($args{type}, $args{prompt}, $args{callback});
+}
+
+sub start_question {
+    my ($prompt, $callback) = @_;
+    BarnOwl::start_edit(type => 'question', prompt => $prompt, callback => sub {
+            my ($text, $success) = @_;
+            $callback->($text) if $success;
+        });
+}
+
+sub start_password {
+    my ($prompt, $callback) = @_;
+    BarnOwl::start_edit(type => 'password', prompt => $prompt, callback => sub {
+            my ($text, $success) = @_;
+            $callback->($text) if $success;
+        });
+}
+
+sub start_edit_win {
+    my ($prompt, $callback) = @_;
+    BarnOwl::start_edit(type => 'edit_win', prompt => $prompt, callback => sub {
+            my ($text, $success) = @_;
+            $callback->($text) if $success;
+        });
+}
 
 =head2 get_data_dir
 
@@ -162,8 +244,9 @@ Adds a file descriptor to C<BarnOwl>'s internal C<select()>
 loop. C<CALLBACK> will be invoked whenever data is available to be
 read from C<FD>.
 
-C<add_dispatch> has been deprecated in favor of C<add_io_dispatch>,
-and is now a wrapper for it called with C<mode> set to C<'r'>.
+C<add_dispatch> has been deprecated in favor of C<AnyEvent>, and is
+now a wrapper for C<add_io_dispatch> called with C<mode> set to
+C<'r'>.
 
 =cut
 
@@ -177,8 +260,7 @@ sub add_dispatch {
 
 Remove a file descriptor previously registered via C<add_dispatch>
 
-C<remove_dispatch> has been deprecated in favor of
-C<remove_io_dispatch>.
+C<remove_dispatch> has been deprecated in favor of C<AnyEvent>.
 
 =cut
 
@@ -193,19 +275,27 @@ whenever C<FD> becomes ready, as specified by <MODE>.
 Only one callback can be registered per FD. If a new callback is
 registered, the old one is removed.
 
+C<add_io_dispatch> has been deprecated in favor of C<AnyEvent>.
+
 =cut
 
+our %_io_dispatches;
+
 sub add_io_dispatch {
     my $fd = shift;
     my $modeStr = shift;
     my $cb = shift;
-    my $mode = 0;
+    my @modes;
 
-    $mode |= 0x1 if ($modeStr =~ /r/i); # Read
-    $mode |= 0x2 if ($modeStr =~ /w/i); # Write
-    if ($mode) {
-        $mode |= 0x4;                  # Exceptional
-        BarnOwl::Internal::add_io_dispatch($fd, $mode, $cb);
+    push @modes, 'r' if $modeStr =~ /r/i; # Read
+    push @modes, 'w' if $modeStr =~ /w/i; # Write
+    if (@modes) {
+	BarnOwl::remove_io_dispatch($fd);
+	for my $mode (@modes) {
+	    push @{$_io_dispatches{$fd}}, AnyEvent->io(fh => $fd,
+						       poll => $mode,
+						       cb => $cb);
+	}
     } else {
         die("Invalid I/O Dispatch mode: $modeStr");
     }
@@ -215,9 +305,19 @@ sub add_io_dispatch {
 
 Remove a file descriptor previously registered via C<add_io_dispatch>
 
+C<remove_io_dispatch> has been deprecated in favor of C<AnyEvent>.
+
+=cut
+
+sub remove_io_dispatch {
+    my $fd = shift;
+    undef $_ foreach @{$_io_dispatches{$fd}};
+    delete $_io_dispatches{$fd};
+}
+
 =head2 create_style NAME OBJECT
 
-Creates a new barnowl style with the given NAME defined by the given
+Creates a new BarnOwl style with the given NAME defined by the given
 object. The object must have a C<description> method which returns a
 string description of the style, and a and C<format_message> method
 which accepts a C<BarnOwl::Message> object and returns a string that
@@ -236,10 +336,15 @@ our $configfile;
 
 our @all_commands;
 
-if(!$configfile && -f $ENV{HOME} . "/.barnowlconf") {
-    $configfile = $ENV{HOME} . "/.barnowlconf";
+if(!$configfile) {
+    if (-f get_config_dir() . "/init.pl") {
+        $configfile = get_config_dir() . "/init.pl";
+    } elsif (-f $ENV{HOME} . "/.barnowlconf") {
+        $configfile = $ENV{HOME} . "/.barnowlconf";
+    } else {
+        $configfile = $ENV{HOME}."/.owlconf";
+    }
 }
-$configfile ||= $ENV{HOME}."/.owlconf";
 
 # populate global variable space for legacy owlconf files
 sub _receive_msg_legacy_wrap {
@@ -548,8 +653,9 @@ Compute the default zephyr signature.
 sub default_zephyr_signature
 {
   my $zsig = getvar('zsig');
-  if (!$zsig) {
-      if (my $zsigproc = getvar('zsigproc')) {
+  if (!defined($zsig) || $zsig eq '') {
+      my $zsigproc = getvar('zsigproc');
+      if (defined($zsigproc) && $zsigproc ne '') {
 	  $zsig = `$zsigproc`;
       } elsif (!defined($zsig = get_zephyr_variable('zwrite-signature'))) {
 	  $zsig = ((getpwuid($<))[6]);
@@ -560,6 +666,25 @@ sub default_zephyr_signature
   return $zsig;
 }
 
+=head3 random_zephyr_signature
+
+Retrieve a random line from ~/.zsigs (except those beginning with '#') 
+and use it as the zephyr signature.
+
+=cut
+
+sub random_zephyr_signature
+{
+    my $zsigfile = "$ENV{'HOME'}/.zsigs";
+    open my $file, '<', $zsigfile or die "Error opening file $zsigfile: $!";
+    my @lines = grep !(/^#/ || /^\s*$/), <$file>;
+    close $file;
+    return '' if !@lines;
+    my $zsig = "$lines[int(rand(scalar @lines))]";
+    chomp $zsig;
+    return $zsig;
+}
+
 # Stub for owl::startup / BarnOwl::startup, so it isn't bound to the
 # startup command. This may be redefined in a user's configfile.
 sub startup
