@@ -78,11 +78,13 @@ our @EXPORT_OK = qw($startup $shutdown
 
 our %EXPORT_TAGS = (all => [@EXPORT_OK]);
 
+use BarnOwl::MainLoopCompatHook;
+
 our $startup = BarnOwl::Hook->new;
 our $shutdown = BarnOwl::Hook->new;
 our $receiveMessage = BarnOwl::Hook->new;
 our $newMessage = BarnOwl::Hook->new;
-our $mainLoop = BarnOwl::Hook->new;
+our $mainLoop = BarnOwl::MainLoopCompatHook->new;
 our $getBuddyList = BarnOwl::Hook->new;
 our $getQuickstart = BarnOwl::Hook->new;
 
@@ -143,7 +145,7 @@ sub _load_owlconf {
     }
 }
 
-# These are the internal hooks called by the barnowl C code, which
+# These are the internal hooks called by the BarnOwl C code, which
 # take care of dispatching to the appropriate perl hooks, and deal
 # with compatibility by calling the old, fixed-name hooks.
 
@@ -161,6 +163,7 @@ sub _startup {
         BarnOwl::error("Can't load BarnOwl::ModuleLoader, loadable module support disabled:\n$@");
     }
     
+    $mainLoop->check_owlconf();
     $startup->run(0);
     BarnOwl::startup() if *BarnOwl::startup{CODE};
 }
@@ -187,11 +190,6 @@ sub _new_msg {
     BarnOwl::new_msg($m) if *BarnOwl::new_msg{CODE};
 }
 
-sub _mainloop_hook {
-    $mainLoop->run;
-    BarnOwl::mainloop_hook() if *BarnOwl::mainloop_hook{CODE};
-}
-
 sub _get_blist {
     my @results = grep defined, $getBuddyList->run;
     s/^\s+|\s+$//sg for (@results);
