@@ -33,6 +33,11 @@ sub should_bold {
     return $m->is_personal && $m->direction eq "in";
 }
 
+sub maybe {
+    my $x = shift;
+    return defined($x) ? $x : "";
+}
+
 sub description {"Default style";}
 
 BarnOwl::create_style("default", "BarnOwl::Style::Default");
@@ -91,14 +96,25 @@ sub chat_header {
         if ( $m->direction eq "out" ) {
             $header = ucfirst $m->type . $personal_context . " sent to " . $m->pretty_recipient;
         } else {
-            $header = ucfirst $m->type . $personal_context . " from " . $m->pretty_sender;
+            $header = ucfirst $m->type . $personal_context . " from ";
+            if(defined($m->auth) && ($m->auth ne "YES")) {
+                $header .= "UNAUTH: ";
+            }
+            $header .= maybe($m->pretty_sender);
         }
     } else {
         $header = $self->humanize($m->context, 1);
         if(defined $m->subcontext) {
             $header .= ' / ' . $self->humanize($m->subcontext, 1);
         }
-        $header .= ' / @b{' . $m->pretty_sender . '}';
+        $header .= ' / ';
+        if(defined($m->auth) && ($m->auth ne "YES")) {
+            $header .= "UNAUTH: ";
+        }
+        $header .= '@b{' . maybe($m->pretty_sender) . '}';
+        if (defined($m->realm) && $m->realm ne BarnOwl::zephyr_getrealm()) {
+            $header .= ' {' . $self->humanize($m->realm, 1) . '}';
+        }
     }
 
     if($m->opcode) {
