@@ -1,17 +1,12 @@
-#include <stdio.h>
-#include <stdlib.h>
-#include <string.h>
-#include <sys/types.h>
-#include <sys/stat.h>
-#include <errno.h>
 #define OWL_PERL
 #include "owl.h"
+#include <stdio.h>
 
 extern XS(boot_BarnOwl);
 extern XS(boot_DynaLoader);
 /* extern XS(boot_DBI); */
 
-static void owl_perl_xs_init(pTHX)
+void owl_perl_xs_init(pTHX) /* noproto */
 {
   const char *file = __FILE__;
   dXSUB_SYS;
@@ -22,7 +17,7 @@ static void owl_perl_xs_init(pTHX)
 }
 
 
-SV *owl_new_sv(const char * str)
+CALLER_OWN SV *owl_new_sv(const char * str)
 {
   SV *ret = newSVpv(str, 0);
   if (is_utf8_string((const U8 *)str, strlen(str))) {
@@ -30,12 +25,12 @@ SV *owl_new_sv(const char * str)
   } else {
     char *escape = owl_escape_highbit(str);
     owl_function_error("Internal error! Non-UTF-8 string encountered:\n%s", escape);
-    owl_free(escape);
+    g_free(escape);
   }
   return ret;
 }
 
-AV *owl_new_av(const owl_list *l, SV *(*to_sv)(const void *))
+CALLER_OWN AV *owl_new_av(const GPtrArray *l, SV *(*to_sv)(const void *))
 {
   AV *ret;
   int i;
@@ -43,18 +38,18 @@ AV *owl_new_av(const owl_list *l, SV *(*to_sv)(const void *))
 
   ret = newAV();
 
-  for (i = 0; i < owl_list_get_size(l); i++) {
-    element = owl_list_get_element(l, i);
+  for (i = 0; i < l->len; i++) {
+    element = l->pdata[i];
     av_push(ret, to_sv(element));
   }
 
   return ret;
 }
 
-HV *owl_new_hv(const owl_dict *d, SV *(*to_sv)(const void *))
+CALLER_OWN HV *owl_new_hv(const owl_dict *d, SV *(*to_sv)(const void *))
 {
   HV *ret;
-  owl_list l;
+  GPtrArray *keys;
   const char *key;
   void *element;
   int i;
@@ -62,24 +57,25 @@ HV *owl_new_hv(const owl_dict *d, SV *(*to_sv)(const void *))
   ret = newHV();
 
   /* TODO: add an iterator-like interface to owl_dict */
-  owl_dict_get_keys(d, &l);
-  for (i = 0; i < owl_list_get_size(&l); i++) {
-    key = owl_list_get_element(&l, i);
+  keys = owl_dict_get_keys(d);
+  for (i = 0; i < keys->len; i++) {
+    key = keys->pdata[i];
     element = owl_dict_find_element(d, key);
     (void)hv_store(ret, key, strlen(key), to_sv(element), 0);
   }
-  owl_list_free_all(&l, owl_free);
+  owl_ptr_array_free(keys, g_free);
 
   return ret;
 }
 
-SV *owl_perlconfig_message2hashref(const owl_message *m)
+CALLER_OWN SV *owl_perlconfig_message2hashref(const owl_message *m)
 {
   HV *h, *stash;
   SV *hr;
   const char *type;
   char *ptr, *utype, *blessas;
-  int i, j;
+  const char *f;
+  int i;
   const owl_pair *pair;
   const owl_filter *wrap;
 
@@ -95,27 +91,28 @@ SV *owl_perlconfig_message2hashref(const owl_message *m)
 #define MSG2H(h,field) (void)hv_store(h, #field, strlen(#field),        \
                                       owl_new_sv(owl_message_get_##field(m)), 0)
 
-  if (owl_message_is_type_zephyr(m)
-      && owl_message_is_direction_in(m)) {
+  if (owl_message_is_type_zephyr(m) && owl_message_is_direction_in(m)) {
     /* Handle zephyr-specific fields... */
-    AV *av_zfields;
-
-    av_zfields = newAV();
-    j=owl_zephyr_get_num_fields(owl_message_get_notice(m));
-    for (i=0; i<j; i++) {
-      ptr=owl_zephyr_get_field_as_utf8(owl_message_get_notice(m), i+1);
-      av_push(av_zfields, owl_new_sv(ptr));
-      owl_free(ptr);
+    AV *av_zfields = newAV();
+    if (owl_message_get_notice(m)) {
+      for (f = owl_zephyr_first_raw_field(owl_message_get_notice(m)); f != NULL;
+           f = owl_zephyr_next_raw_field(owl_message_get_notice(m), f)) {
+        ptr = owl_zephyr_field_as_utf8(owl_message_get_notice(m), f);
+        av_push(av_zfields, owl_new_sv(ptr));
+        g_free(ptr);
+      }
+      (void)hv_store(h, "auth", strlen("auth"),
+                     owl_new_sv(owl_zephyr_get_authstr(owl_message_get_notice(m))), 0);
+    } else {
+      /* Incoming zephyrs without a ZNotice_t are pseudo-logins. To appease
+       * existing styles, put in bogus 'auth' and 'fields' keys. */
+      (void)hv_store(h, "auth", strlen("auth"), owl_new_sv("NO"), 0);
     }
     (void)hv_store(h, "fields", strlen("fields"), newRV_noinc((SV*)av_zfields), 0);
-
-    (void)hv_store(h, "auth", strlen("auth"), 
-                   owl_new_sv(owl_zephyr_get_authstr(owl_message_get_notice(m))),0);
   }
 
-  j=owl_list_get_size(&(m->attributes));
-  for(i=0; i<j; i++) {
-    pair=owl_list_get_element(&(m->attributes), i);
+  for (i = 0; i < m->attributes->len; i++) {
+    pair = m->attributes->pdata[i];
     (void)hv_store(h, owl_pair_get_key(pair), strlen(owl_pair_get_key(pair)),
                    owl_new_sv(owl_pair_get_value(pair)),0);
   }
@@ -147,9 +144,9 @@ SV *owl_perlconfig_message2hashref(const owl_message *m)
 
   type = owl_message_get_type(m);
   if(!type || !*type) type = "generic";
-  utype = owl_strdup(type);
+  utype = g_strdup(type);
   utype[0] = toupper(type[0]);
-  blessas = owl_sprintf("BarnOwl::Message::%s", utype);
+  blessas = g_strdup_printf("BarnOwl::Message::%s", utype);
 
   hr = newRV_noinc((SV*)h);
   stash =  gv_stashpv(blessas,0);
@@ -158,12 +155,12 @@ SV *owl_perlconfig_message2hashref(const owl_message *m)
     stash = gv_stashpv("BarnOwl::Message", 1);
   }
   hr = sv_bless(hr,stash);
-  owl_free(utype);
-  owl_free(blessas);
+  g_free(utype);
+  g_free(blessas);
   return hr;
 }
 
-SV *owl_perlconfig_curmessage2hashref(void)
+CALLER_OWN SV *owl_perlconfig_curmessage2hashref(void)
 {
   int curmsg;
   const owl_view *v;
@@ -178,24 +175,22 @@ SV *owl_perlconfig_curmessage2hashref(void)
 /* XXX TODO: Messages should round-trip properly between
    message2hashref and hashref2message. Currently we lose
    zephyr-specific properties stored in the ZNotice_t
-
-   This has been somewhat addressed, but is still not lossless.
  */
-owl_message * owl_perlconfig_hashref2message(SV *msg)
+CALLER_OWN owl_message *owl_perlconfig_hashref2message(SV *msg)
 {
   owl_message * m;
   HE * ent;
-  I32 count, len;
+  I32 len;
   const char *key,*val;
   HV * hash;
   struct tm tm;
 
   hash = (HV*)SvRV(msg);
 
-  m = owl_malloc(sizeof(owl_message));
+  m = g_new(owl_message, 1);
   owl_message_init(m);
 
-  count = hv_iterinit(hash);
+  hv_iterinit(hash);
   while((ent = hv_iternext(hash))) {
     key = hv_iterkey(ent, &len);
     val = SvPV_nolen(hv_iterval(hash, ent));
@@ -213,7 +208,8 @@ owl_message * owl_perlconfig_hashref2message(SV *msg)
     } else if (!strcmp(key, "zwriteline")) {
       owl_message_set_zwriteline(m, val);
     } else if (!strcmp(key, "time")) {
-      m->timestr = owl_strdup(val);
+      g_free(m->timestr);
+      m->timestr = g_strdup(val);
       strptime(val, "%a %b %d %T %Y", &tm);
       m->time = mktime(&tm);
     } else {
@@ -224,31 +220,12 @@ owl_message * owl_perlconfig_hashref2message(SV *msg)
     if(!owl_message_get_attribute_value(m, "adminheader"))
       owl_message_set_attribute(m, "adminheader", "");
   }
-#ifdef HAVE_LIBZEPHYR
-  if (owl_message_is_type_zephyr(m)) {
-    ZNotice_t *n = &(m->notice);
-    n->z_kind = ACKED;
-    n->z_port = 0;
-    n->z_auth = ZAUTH_NO;
-    n->z_checked_auth = 0;
-    n->z_class = zstr(owl_message_get_class(m));
-    n->z_class_inst = zstr(owl_message_get_instance(m));
-    n->z_opcode = zstr(owl_message_get_opcode(m));
-    n->z_sender = zstr(owl_message_get_sender(m));
-    n->z_recipient = zstr(owl_message_get_recipient(m));
-    n->z_default_format = zstr("[zephyr created from perl]");
-    n->z_multinotice = zstr("[zephyr created from perl]");
-    n->z_num_other_fields = 0;
-    n->z_message = owl_sprintf("%s%c%s", owl_message_get_zsig(m), '\0', owl_message_get_body(m));
-    n->z_message_len = strlen(owl_message_get_zsig(m)) + strlen(owl_message_get_body(m)) + 1;
-  }
-#endif
   return m;
 }
 
 /* Calls in a scalar context, passing it a hash reference.
    If return value is non-null, caller must free. */
-char *owl_perlconfig_call_with_message(const char *subname, const owl_message *m)
+CALLER_OWN char *owl_perlconfig_call_with_message(const char *subname, const owl_message *m)
 {
   dSP ;
   int count;
@@ -281,7 +258,7 @@ char *owl_perlconfig_call_with_message(const char *subname, const owl_message *m
   srv = POPs;
 
   if (srv) {
-    out = owl_strdup(SvPV_nolen(srv));
+    out = g_strdup(SvPV_nolen(srv));
   } else {
     out = NULL;
   }
@@ -297,7 +274,7 @@ char *owl_perlconfig_call_with_message(const char *subname, const owl_message *m
 /* Calls a method on a perl object representing a message.
    If the return value is non-null, the caller must free it.
  */
-char * owl_perlconfig_message_call_method(const owl_message *m, const char *method, int argc, const char ** argv)
+CALLER_OWN char *owl_perlconfig_message_call_method(const owl_message *m, const char *method, int argc, const char **argv)
 {
   dSP;
   unsigned int count, i;
@@ -334,7 +311,7 @@ char * owl_perlconfig_message_call_method(const owl_message *m, const char *meth
   srv = POPs;
 
   if (srv) {
-    out = owl_strdup(SvPV_nolen(srv));
+    out = g_strdup(SvPV_nolen(srv));
   } else {
     out = NULL;
   }
@@ -346,8 +323,8 @@ char * owl_perlconfig_message_call_method(const owl_message *m, const char *meth
   return out;
 }
 
-
-char *owl_perlconfig_initperl(const char * file, int *Pargc, char ***Pargv, char *** Penv)
+/* caller must free result, if not NULL */
+CALLER_OWN char *owl_perlconfig_initperl(const char *file, int *Pargc, char ***Pargv, char ***Penv)
 {
   int ret;
   PerlInterpreter *p;
@@ -362,18 +339,20 @@ char *owl_perlconfig_initperl(const char * file, int *Pargc, char ***Pargv, char
   owl_global_set_perlinterp(&g, p);
   perl_construct(p);
 
+  PL_exit_flags |= PERL_EXIT_DESTRUCT_END;
+
   owl_global_set_no_have_config(&g);
 
   ret=perl_parse(p, owl_perl_xs_init, 2, (char **)args, NULL);
   if (ret || SvTRUE(ERRSV)) {
-    err=owl_strdup(SvPV_nolen(ERRSV));
+    err=g_strdup(SvPV_nolen(ERRSV));
     sv_setsv(ERRSV, &PL_sv_undef);     /* and clear the error */
     return(err);
   }
 
   ret=perl_run(p);
   if (ret || SvTRUE(ERRSV)) {
-    err=owl_strdup(SvPV_nolen(ERRSV));
+    err=g_strdup(SvPV_nolen(ERRSV));
     sv_setsv(ERRSV, &PL_sv_undef);     /* and clear the error */
     return(err);
   }
@@ -399,17 +378,19 @@ char *owl_perlconfig_initperl(const char * file, int *Pargc, char ***Pargv, char
     sv_setpv(cfg, file);
   }
 
+  sv_setpv(get_sv("BarnOwl::VERSION", TRUE), OWL_VERSION_STRING);
+
   /* Add the system lib path to @INC */
   inc = get_av("INC", 0);
-  path = owl_sprintf("%s/lib", owl_get_datadir());
+  path = g_build_filename(owl_get_datadir(), "lib", NULL);
   av_unshift(inc, 1);
   av_store(inc, 0, owl_new_sv(path));
-  owl_free(path);
+  g_free(path);
 
   eval_pv("use BarnOwl;", FALSE);
 
   if (SvTRUE(ERRSV)) {
-    err=owl_strdup(SvPV_nolen(ERRSV));
+    err=g_strdup(SvPV_nolen(ERRSV));
     sv_setsv (ERRSV, &PL_sv_undef);     /* and clear the error */
     return(err);
   }
@@ -429,7 +410,7 @@ int owl_perlconfig_is_function(const char *fn) {
 }
 
 /* caller is responsible for freeing returned string */
-char *owl_perlconfig_execute(const char *line)
+CALLER_OWN char *owl_perlconfig_execute(const char *line)
 {
   STRLEN len;
   SV *response;
@@ -447,7 +428,7 @@ char *owl_perlconfig_execute(const char *line)
     sv_setsv (ERRSV, &PL_sv_undef);     /* and clear the error */
   }
 
-  out = owl_strdup(SvPV(response, len));
+  out = g_strdup(SvPV(response, len));
   FREETMPS;
   LEAVE;
 
@@ -461,7 +442,7 @@ void owl_perlconfig_getmsg(const owl_message *m, const char *subname)
     ptr = owl_perlconfig_call_with_message(subname?subname
                                            :"BarnOwl::_receive_msg_legacy_wrap", m);
   }
-  if (ptr) owl_free(ptr);
+  g_free(ptr);
 }
 
 /* Called on all new messages; receivemsg is only called on incoming ones */
@@ -472,7 +453,7 @@ void owl_perlconfig_newmsg(const owl_message *m, const char *subname)
     ptr = owl_perlconfig_call_with_message(subname?subname
                                            :"BarnOwl::Hooks::_new_msg", m);
   }
-  if (ptr) owl_free(ptr);
+  g_free(ptr);
 }
 
 void owl_perlconfig_new_command(const char *name)
@@ -486,7 +467,7 @@ void owl_perlconfig_new_command(const char *name)
   XPUSHs(sv_2mortal(owl_new_sv(name)));
   PUTBACK;
 
-  call_pv("BarnOwl::Hooks::_new_command", G_SCALAR|G_VOID);
+  call_pv("BarnOwl::Hooks::_new_command", G_VOID|G_EVAL);
 
   SPAGAIN;
 
@@ -498,7 +479,8 @@ void owl_perlconfig_new_command(const char *name)
   LEAVE;
 }
 
-char *owl_perlconfig_perlcmd(const owl_cmd *cmd, int argc, const char *const *argv)
+/* caller must free the result */
+CALLER_OWN char *owl_perlconfig_perlcmd(const owl_cmd *cmd, int argc, const char *const *argv)
 {
   int i, count;
   char * ret = NULL;
@@ -526,7 +508,7 @@ char *owl_perlconfig_perlcmd(const owl_cmd *cmd, int argc, const char *const *ar
       croak("Perl command %s returned more than one value!", cmd->name);
     rv = POPs;
     if(SvTRUE(rv)) {
-      ret = owl_strdup(SvPV_nolen(rv));
+      ret = g_strdup(SvPV_nolen(rv));
     }
   }
 
@@ -536,17 +518,12 @@ char *owl_perlconfig_perlcmd(const owl_cmd *cmd, int argc, const char *const *ar
   return ret;
 }
 
-void owl_perlconfig_cmd_free(owl_cmd *cmd)
+void owl_perlconfig_cmd_cleanup(owl_cmd *cmd)
 {
   SvREFCNT_dec(cmd->cmd_perl);
 }
 
-void owl_perlconfig_io_dispatch_destroy(const owl_io_dispatch *d)
-{
-  SvREFCNT_dec(d->data);
-}
-
-void owl_perlconfig_edit_callback(owl_editwin *e)
+void owl_perlconfig_edit_callback(owl_editwin *e, bool success)
 {
   SV *cb = owl_editwin_get_cbdata(e);
   SV *text;
@@ -563,6 +540,7 @@ void owl_perlconfig_edit_callback(owl_editwin *e)
 
   PUSHMARK(SP);
   XPUSHs(sv_2mortal(text));
+  XPUSHs(sv_2mortal(newSViv(success)));
   PUTBACK;
   
   call_sv(cb, G_DISCARD|G_EVAL);
@@ -573,82 +551,10 @@ void owl_perlconfig_edit_callback(owl_editwin *e)
 
   FREETMPS;
   LEAVE;
-
-  SvREFCNT_dec(cb);
-  owl_editwin_set_cbdata(e, NULL);
 }
 
-void owl_perlconfig_mainloop(owl_timer *t, void *data)
+void owl_perlconfig_dec_refcnt(void *data)
 {
-  dSP;
-  if (!owl_perlconfig_is_function("BarnOwl::Hooks::_mainloop_hook"))
-    return;
-  PUSHMARK(SP) ;
-  call_pv("BarnOwl::Hooks::_mainloop_hook", G_DISCARD|G_EVAL);
-  if(SvTRUE(ERRSV)) {
-    owl_function_error("%s", SvPV_nolen(ERRSV));
-  }
-  return;
-}
-
-void owl_perlconfig_io_dispatch(const owl_io_dispatch *d, void *data)
-{
-  SV *cb = data;
-  dSP;
-  if(cb == NULL) {
-    owl_function_error("Perl callback is NULL!");
-    return;
-  }
-
-  ENTER;
-  SAVETMPS;
-
-  PUSHMARK(SP);
-  PUTBACK;
-
-  call_sv(cb, G_DISCARD|G_EVAL);
-
-  if(SvTRUE(ERRSV)) {
-    owl_function_error("%s", SvPV_nolen(ERRSV));
-  }
-
-  FREETMPS;
-  LEAVE;
-}
-
-void owl_perlconfig_perl_timer(owl_timer *t, void *data)
-{
-  dSP;
-  SV *obj = data;
-
-  if(!SvROK(obj)) {
-    return;
-  }
-
-  ENTER;
-  SAVETMPS;
-
-  PUSHMARK(SP);
-  XPUSHs(obj);
-  PUTBACK;
-
-  call_method("do_callback", G_DISCARD|G_EVAL);
-
-  SPAGAIN;
-
-  if (SvTRUE(ERRSV)) {
-    owl_function_error("Error in calback: '%s'", SvPV_nolen(ERRSV));
-    sv_setsv (ERRSV, &PL_sv_undef);
-  }
-
-  PUTBACK;
-  FREETMPS;
-  LEAVE;
-}
-
-void owl_perlconfig_perl_timer_destroy(owl_timer *t)
-{
-  if(SvOK((SV*)t->data)) {
-    SvREFCNT_dec((SV*)t->data);
-  }
+  SV *v = data;
+  SvREFCNT_dec(v);
 }
