@@ -1,4 +1,3 @@
-#include <string.h>
 #include "owl.h"
 
 void owl_regex_init(owl_regex *re)
@@ -10,10 +9,11 @@ void owl_regex_init(owl_regex *re)
 int owl_regex_create(owl_regex *re, const char *string)
 {
   int ret;
-  char buff1[LINE];
+  size_t errbuf_size;
+  char *errbuf;
   const char *ptr;
   
-  re->string=owl_strdup(string);
+  re->string=g_strdup(string);
 
   ptr=string;
   re->negate=0;
@@ -25,9 +25,12 @@ int owl_regex_create(owl_regex *re, const char *string)
   /* set the regex */
   ret=regcomp(&(re->re), ptr, REG_EXTENDED|REG_ICASE);
   if (ret) {
-    regerror(ret, NULL, buff1, LINE);
-    owl_function_makemsg("Error in regular expression: %s", buff1);
-    owl_free(re->string);
+    errbuf_size = regerror(ret, NULL, NULL, 0);
+    errbuf = g_new(char, errbuf_size);
+    regerror(ret, NULL, errbuf, errbuf_size);
+    owl_function_error("Error in regular expression: %s", errbuf);
+    g_free(errbuf);
+    g_free(re->string);
     re->string=NULL;
     return(-1);
   }
@@ -38,11 +41,12 @@ int owl_regex_create(owl_regex *re, const char *string)
 int owl_regex_create_quoted(owl_regex *re, const char *string)
 {
   char *quoted;
+  int ret;
   
-  quoted=owl_text_quote(string, OWL_REGEX_QUOTECHARS, OWL_REGEX_QUOTEWITH);
-  owl_regex_create(re, quoted);
-  owl_free(quoted);
-  return(0);
+  quoted = owl_text_quote(string, OWL_REGEX_QUOTECHARS, OWL_REGEX_QUOTEWITH);
+  ret = owl_regex_create(re, quoted);
+  g_free(quoted);
+  return ret;
 }
 
 int owl_regex_compare(const owl_regex *re, const char *string, int *start, int *end)
@@ -86,7 +90,7 @@ void owl_regex_copy(const owl_regex *a, owl_regex *b)
 void owl_regex_cleanup(owl_regex *re)
 {
     if (re->string) {
-        owl_free(re->string);
+        g_free(re->string);
         regfree(&(re->re));
     }
 }
